import re
from typing import Any, Callable, List, Match, Optional, Pattern, Union

MatchEvaluator = Callable[[Any], str]


def _options_to_flags(options: int) -> int:
    """Convert .NET options to Python flags."""
    flags = 256
    if options & 1:
        flags |= re.IGNORECASE
    if options & 2:
        flags |= re.MULTILINE
    if options & 16:
        flags |= re.S
    return flags


def create(pattern: str, options: int = 0) -> Pattern[str]:
    # raw_pattern = pattern.encode("unicode_escape").decode("utf-8")
    flags = _options_to_flags(options)
    return re.compile(pattern, flags=flags)


def escape(string: str) -> str:
    return re.escape(string)


def unescape(string: str) -> str:
    return re.sub(r"\\(.)", r"\1", string)


def match(
    reg: Union[Pattern[str], str], input: str, start_at: int = 0
) -> Optional[Match[str]]:
    if isinstance(reg, str):
        flags = _options_to_flags(start_at)
        return re.search(input, reg, flags)
    return reg.search(input, pos=start_at)


def matches(reg: Pattern[str], input: str, start_at: int = 0) -> List[Match[str]]:
    if isinstance(reg, str):
        flags = _options_to_flags(start_at)
        return re.findall(input, reg, flags=flags)

    return reg.findall(input, pos=start_at)


def is_match(reg: Union[Pattern[str], str], input: str, start_at: int = 0) -> bool:
    if isinstance(reg, str):
        # Note: input is the pattern here
        flags = _options_to_flags(start_at)
        return re.search(input, reg, flags=flags) is not None

    return reg.search(input, pos=start_at) is not None


def groups(m: Match[str]) -> List[str]:
    # .NET adds the whole capture as group 0
    g = list(m.groups())
    g.insert(0, m.string)
    return g


def options(reg: Pattern[str]) -> int:
    options = 256  # ECMAScript
    if reg.flags & re.IGNORECASE:
        options |= 1
    if reg.flags & re.MULTILINE:
        options |= 2
    if reg.flags & re.S:
        options |= 16
    return options


def replace(
    reg: Union[str, Pattern[str]],
    input: str,
    replacement: Union[str, MatchEvaluator],
    limit: Optional[int] = None,
    offset: int = 0,
) -> str:
    if isinstance(replacement, str):
        # Rewrite placeholders from .NET e.g $1 to \g<1> Python syntax
        replacement = re.sub(pattern=r"\$(\d+)", repl=r"\\g<\g<1>>", string=replacement)

    if isinstance(reg, str):
        # print("reg, replacement, input=", reg, replacement, input)
        return re.sub(input, replacement, reg, count=limit or 0)
    else:
        return input[:offset] + reg.sub(replacement, input[offset:], count=limit or 0)


def split(
    reg: Union[str, Pattern[str]],
    input: str,
    limit: Optional[int] = None,
    offset: int = 0,
) -> List[str]:
    if isinstance(reg, str):
        return re.split(input, reg, maxsplit=limit or 0)

    input = input[offset:]
    # print("reg, input: ", reg, input, limit)
    return reg.split(input, maxsplit=limit or 0)[:limit]


__all__ = [
    "escape",
    "is_match",
    "match",
    "matches",
    "create",
    "options",
    "replace",
    "split",
    "unescape",
]



from __future__ import annotations
from typing import (TypeVar, Generic, Optional, Any, Tuple, Callable, List)
from .array import (fill, map as map_2)
from .list import (FSharpList, cons, empty as empty_1, fold as fold_1, is_empty as is_empty_1, tail, head, of_array_with_tail, singleton)
from .option import (value as value_1, some)
from .reflection import (TypeInfo, class_type, option_type, list_type, bool_type, record_type)
from .seq import (unfold, map as map_1, compare_with, iterate as iterate_1, pick as pick_1, try_pick as try_pick_1)
from .string import (join, format)
from .types import (Record, FSharpRef)
from .util import (IComparer, IEnumerator, IEnumerable, is_array_like, get_enumerator, equals, to_iterator, compare, ignore, ICollection, structural_hash)

_KEY = TypeVar("_KEY")

_VALUE = TypeVar("_VALUE")

_A_ = TypeVar("_A_")

_B_ = TypeVar("_B_")

_C_ = TypeVar("_C_")

_RESULT = TypeVar("_RESULT")

_Z = TypeVar("_Z")

_B = TypeVar("_B")

_STATE = TypeVar("_STATE")

_T = TypeVar("_T")

_K = TypeVar("_K")

_V = TypeVar("_V")

def expr_266(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return class_type("Map.MapTreeLeaf`2", [gen0, gen1], MapTreeLeaf_2)


class MapTreeLeaf_2(Generic[_KEY, _VALUE]):
    def __init__(self, k: _KEY, v: _VALUE) -> None:
        self.k = k
        self.v = v
    

MapTreeLeaf_2_reflection = expr_266

def MapTreeLeaf_2__ctor_5BDDA1(k: _KEY, v: _VALUE) -> MapTreeLeaf_2[_KEY, _VALUE]:
    return MapTreeLeaf_2(k, v)


def MapTreeLeaf_2__get_Key(_: MapTreeLeaf_2[_KEY, _VALUE]) -> _KEY:
    return _.k


def MapTreeLeaf_2__get_Value(_: MapTreeLeaf_2[_KEY, _VALUE]) -> _VALUE:
    return _.v


def expr_267(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return class_type("Map.MapTreeNode`2", [gen0, gen1], MapTreeNode_2, MapTreeLeaf_2_reflection(gen0, gen1))


class MapTreeNode_2(MapTreeLeaf_2, Generic[_KEY, _VALUE]):
    def __init__(self, k: _KEY, v: _VALUE, left: Optional[MapTreeLeaf_2[_KEY, _VALUE]], right: Optional[MapTreeLeaf_2[_KEY, _VALUE]], h: int) -> None:
        super().__init__(k, v)
        self.left = left
        self.right = right
        self.h = h or 0
    

MapTreeNode_2_reflection = expr_267

def MapTreeNode_2__ctor_Z39DE9543(k: _KEY, v: _VALUE, left: Optional[MapTreeLeaf_2[_KEY, _VALUE]], right: Optional[MapTreeLeaf_2[_KEY, _VALUE]], h: int) -> MapTreeNode_2[_KEY, _VALUE]:
    return MapTreeNode_2(k, v, left, right, h)


def MapTreeNode_2__get_Left(_: MapTreeNode_2[_KEY, _VALUE]) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    return _.left


def MapTreeNode_2__get_Right(_: MapTreeNode_2[_KEY, _VALUE]) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    return _.right


def MapTreeNode_2__get_Height(_: MapTreeNode_2[_KEY, _VALUE]) -> int:
    return _.h


def MapTreeModule_empty() -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    return None


def MapTreeModule_sizeAux(acc_mut: int, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> int:
    while True:
        (acc, m) = (acc_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                acc_mut = MapTreeModule_sizeAux(acc + 1, MapTreeNode_2__get_Left(m2))
                m_mut = MapTreeNode_2__get_Right(m2)
                continue
            
            else: 
                return acc + 1
            
        
        else: 
            return acc
        
        break


def MapTreeModule_size(x: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> int:
    return MapTreeModule_sizeAux(0, x)


def MapTreeModule_mk(l: Optional[MapTreeLeaf_2[_KEY, _VALUE]], k: _KEY, v: _VALUE, r: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    hl : int
    m : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = l
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        hl = MapTreeNode_2__get_Height(m2) if isinstance(m2, MapTreeNode_2) else 1
    
    else: 
        hl = 0
    
    hr : int
    m_1 : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = r
    if m_1 is not None:
        m2_1 : MapTreeLeaf_2[_KEY, _VALUE] = m_1
        hr = MapTreeNode_2__get_Height(m2_1) if isinstance(m2_1, MapTreeNode_2) else 1
    
    else: 
        hr = 0
    
    m_2 : int = (hr if (hl < hr) else hl) or 0
    if m_2 == 0:
        return MapTreeLeaf_2__ctor_5BDDA1(k, v)
    
    else: 
        return MapTreeNode_2__ctor_Z39DE9543(k, v, l, r, m_2 + 1)
    


def MapTreeModule_rebalance(t1: Optional[MapTreeLeaf_2[_KEY, _VALUE]], k: _KEY, v: _VALUE, t2: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    t1h : int
    m : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = t1
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        t1h = MapTreeNode_2__get_Height(m2) if isinstance(m2, MapTreeNode_2) else 1
    
    else: 
        t1h = 0
    
    t2h : int
    m_1 : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = t2
    if m_1 is not None:
        m2_1 : MapTreeLeaf_2[_KEY, _VALUE] = m_1
        t2h = MapTreeNode_2__get_Height(m2_1) if isinstance(m2_1, MapTreeNode_2) else 1
    
    else: 
        t2h = 0
    
    if t2h > (t1h + 2):
        match_value : MapTreeLeaf_2[_KEY, _VALUE] = value_1(t2)
        if isinstance(match_value, MapTreeNode_2):
            def arrow_268(t1: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=t1, k: _KEY=k, v: _VALUE=v, t2: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=t2) -> int:
                m_2 : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = MapTreeNode_2__get_Left(match_value)
                if m_2 is not None:
                    m2_2 : MapTreeLeaf_2[_KEY, _VALUE] = m_2
                    return MapTreeNode_2__get_Height(m2_2) if isinstance(m2_2, MapTreeNode_2) else 1
                
                else: 
                    return 0
                
            
            if arrow_268() > (t1h + 1):
                match_value_1 : MapTreeLeaf_2[_KEY, _VALUE] = value_1(MapTreeNode_2__get_Left(match_value))
                if isinstance(match_value_1, MapTreeNode_2):
                    return MapTreeModule_mk(MapTreeModule_mk(t1, k, v, MapTreeNode_2__get_Left(match_value_1)), MapTreeLeaf_2__get_Key(match_value_1), MapTreeLeaf_2__get_Value(match_value_1), MapTreeModule_mk(MapTreeNode_2__get_Right(match_value_1), MapTreeLeaf_2__get_Key(match_value), MapTreeLeaf_2__get_Value(match_value), MapTreeNode_2__get_Right(match_value)))
                
                else: 
                    raise Exception("internal error: Map.rebalance")
                
            
            else: 
                return MapTreeModule_mk(MapTreeModule_mk(t1, k, v, MapTreeNode_2__get_Left(match_value)), MapTreeLeaf_2__get_Key(match_value), MapTreeLeaf_2__get_Value(match_value), MapTreeNode_2__get_Right(match_value))
            
        
        else: 
            raise Exception("internal error: Map.rebalance")
        
    
    elif t1h > (t2h + 2):
        match_value_2 : MapTreeLeaf_2[_KEY, _VALUE] = value_1(t1)
        if isinstance(match_value_2, MapTreeNode_2):
            def arrow_269(t1: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=t1, k: _KEY=k, v: _VALUE=v, t2: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=t2) -> int:
                m_3 : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = MapTreeNode_2__get_Right(match_value_2)
                if m_3 is not None:
                    m2_3 : MapTreeLeaf_2[_KEY, _VALUE] = m_3
                    return MapTreeNode_2__get_Height(m2_3) if isinstance(m2_3, MapTreeNode_2) else 1
                
                else: 
                    return 0
                
            
            if arrow_269() > (t2h + 1):
                match_value_3 : MapTreeLeaf_2[_KEY, _VALUE] = value_1(MapTreeNode_2__get_Right(match_value_2))
                if isinstance(match_value_3, MapTreeNode_2):
                    return MapTreeModule_mk(MapTreeModule_mk(MapTreeNode_2__get_Left(match_value_2), MapTreeLeaf_2__get_Key(match_value_2), MapTreeLeaf_2__get_Value(match_value_2), MapTreeNode_2__get_Left(match_value_3)), MapTreeLeaf_2__get_Key(match_value_3), MapTreeLeaf_2__get_Value(match_value_3), MapTreeModule_mk(MapTreeNode_2__get_Right(match_value_3), k, v, t2))
                
                else: 
                    raise Exception("internal error: Map.rebalance")
                
            
            else: 
                return MapTreeModule_mk(MapTreeNode_2__get_Left(match_value_2), MapTreeLeaf_2__get_Key(match_value_2), MapTreeLeaf_2__get_Value(match_value_2), MapTreeModule_mk(MapTreeNode_2__get_Right(match_value_2), k, v, t2))
            
        
        else: 
            raise Exception("internal error: Map.rebalance")
        
    
    else: 
        return MapTreeModule_mk(t1, k, v, t2)
    


def MapTreeModule_add(comparer: IComparer[_KEY], k: _KEY, v: _VALUE, m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        c : int = comparer.Compare(k, MapTreeLeaf_2__get_Key(m2)) or 0
        if isinstance(m2, MapTreeNode_2):
            if c < 0:
                return MapTreeModule_rebalance(MapTreeModule_add(comparer, k, v, MapTreeNode_2__get_Left(m2)), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeNode_2__get_Right(m2))
            
            elif c == 0:
                return MapTreeNode_2__ctor_Z39DE9543(k, v, MapTreeNode_2__get_Left(m2), MapTreeNode_2__get_Right(m2), MapTreeNode_2__get_Height(m2))
            
            else: 
                return MapTreeModule_rebalance(MapTreeNode_2__get_Left(m2), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeModule_add(comparer, k, v, MapTreeNode_2__get_Right(m2)))
            
        
        elif c < 0:
            return MapTreeNode_2__ctor_Z39DE9543(k, v, MapTreeModule_empty(), m, 2)
        
        elif c == 0:
            return MapTreeLeaf_2__ctor_5BDDA1(k, v)
        
        else: 
            return MapTreeNode_2__ctor_Z39DE9543(k, v, m, MapTreeModule_empty(), 2)
        
    
    else: 
        return MapTreeLeaf_2__ctor_5BDDA1(k, v)
    


def MapTreeModule_tryFind(comparer_mut: IComparer[_KEY], k_mut: _KEY, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> Optional[_VALUE]:
    while True:
        (comparer, k, m) = (comparer_mut, k_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            c : int = comparer.Compare(k, MapTreeLeaf_2__get_Key(m2)) or 0
            if c == 0:
                return some(MapTreeLeaf_2__get_Value(m2))
            
            elif isinstance(m2, MapTreeNode_2):
                comparer_mut = comparer
                k_mut = k
                m_mut = MapTreeNode_2__get_Left(m2) if (c < 0) else MapTreeNode_2__get_Right(m2)
                continue
            
            else: 
                return None
            
        
        else: 
            return None
        
        break


def MapTreeModule_find(comparer: IComparer[_KEY], k: _KEY, m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> _VALUE:
    match_value : Optional[_VALUE] = MapTreeModule_tryFind(comparer, k, m)
    if match_value is None:
        raise Exception()
    
    else: 
        return value_1(match_value)
    


def MapTreeModule_partition1(comparer: IComparer[_KEY], f: Any, k: _KEY, v: _A_, acc1: Optional[MapTreeLeaf_2[_KEY, _A_]]=None, acc2: Optional[MapTreeLeaf_2[_KEY, _A_]]=None) -> Tuple[Optional[MapTreeLeaf_2[_KEY, _A_]], Optional[MapTreeLeaf_2[_KEY, _A_]]]:
    if f(k, v):
        return (MapTreeModule_add(comparer, k, v, acc1), acc2)
    
    else: 
        return (acc1, MapTreeModule_add(comparer, k, v, acc2))
    


def MapTreeModule_partitionAux(comparer_mut: IComparer[_KEY], f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]], acc_0_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]], acc_1_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> Tuple[Optional[MapTreeLeaf_2[_KEY, _VALUE]], Optional[MapTreeLeaf_2[_KEY, _VALUE]]]:
    while True:
        (comparer, f, m, acc_0, acc_1) = (comparer_mut, f_mut, m_mut, acc_0_mut, acc_1_mut)
        acc : Tuple[Optional[MapTreeLeaf_2[_KEY, _VALUE]], Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = (acc_0, acc_1)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                acc_2 : Tuple[Optional[MapTreeLeaf_2[_KEY, _VALUE]], Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = MapTreeModule_partitionAux(comparer, f, MapTreeNode_2__get_Right(m2), acc[0], acc[1])
                acc_3 : Tuple[Optional[MapTreeLeaf_2[_KEY, _VALUE]], Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = MapTreeModule_partition1(comparer, f, MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), acc_2[0], acc_2[1])
                comparer_mut = comparer
                f_mut = f
                m_mut = MapTreeNode_2__get_Left(m2)
                acc_0_mut = acc_3[0]
                acc_1_mut = acc_3[1]
                continue
            
            else: 
                return MapTreeModule_partition1(comparer, f, MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), acc[0], acc[1])
            
        
        else: 
            return acc
        
        break


def MapTreeModule_partition(comparer: IComparer[_KEY], f: Callable[[_KEY, _A_], bool], m: Optional[MapTreeLeaf_2[_KEY, _A_]]=None) -> Tuple[Optional[MapTreeLeaf_2[_KEY, _A_]], Optional[MapTreeLeaf_2[_KEY, _A_]]]:
    return MapTreeModule_partitionAux(comparer, f, m, MapTreeModule_empty(), MapTreeModule_empty())


def MapTreeModule_filter1(comparer: IComparer[_KEY], f: Any, k: _KEY, v: _A_, acc: Optional[MapTreeLeaf_2[_KEY, _A_]]=None) -> Optional[MapTreeLeaf_2[_KEY, _A_]]:
    if f(k, v):
        return MapTreeModule_add(comparer, k, v, acc)
    
    else: 
        return acc
    


def MapTreeModule_filterAux(comparer_mut: IComparer[_KEY], f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]], acc_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    while True:
        (comparer, f, m, acc) = (comparer_mut, f_mut, m_mut, acc_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                acc_1 : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = MapTreeModule_filterAux(comparer, f, MapTreeNode_2__get_Left(m2), acc)
                acc_2 : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = MapTreeModule_filter1(comparer, f, MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), acc_1)
                comparer_mut = comparer
                f_mut = f
                m_mut = MapTreeNode_2__get_Right(m2)
                acc_mut = acc_2
                continue
            
            else: 
                return MapTreeModule_filter1(comparer, f, MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), acc)
            
        
        else: 
            return acc
        
        break


def MapTreeModule_filter(comparer: IComparer[_KEY], f: Callable[[_KEY, _A_], bool], m: Optional[MapTreeLeaf_2[_KEY, _A_]]=None) -> Optional[MapTreeLeaf_2[_KEY, _A_]]:
    return MapTreeModule_filterAux(comparer, f, m, MapTreeModule_empty())


def MapTreeModule_spliceOutSuccessor(m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Tuple[_KEY, _VALUE, Optional[MapTreeLeaf_2[_KEY, _VALUE]]]:
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        if isinstance(m2, MapTreeNode_2):
            if MapTreeNode_2__get_Left(m2) is None:
                return (MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeNode_2__get_Right(m2))
            
            else: 
                pattern_input : Tuple[_KEY, _VALUE, Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = MapTreeModule_spliceOutSuccessor(MapTreeNode_2__get_Left(m2))
                return (pattern_input[0], pattern_input[1], MapTreeModule_mk(pattern_input[2], MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeNode_2__get_Right(m2)))
            
        
        else: 
            return (MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeModule_empty())
        
    
    else: 
        raise Exception("internal error: Map.spliceOutSuccessor")
    


def MapTreeModule_remove(comparer: IComparer[_KEY], k: _KEY, m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        c : int = comparer.Compare(k, MapTreeLeaf_2__get_Key(m2)) or 0
        if isinstance(m2, MapTreeNode_2):
            if c < 0:
                return MapTreeModule_rebalance(MapTreeModule_remove(comparer, k, MapTreeNode_2__get_Left(m2)), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeNode_2__get_Right(m2))
            
            elif c == 0:
                if MapTreeNode_2__get_Left(m2) is None:
                    return MapTreeNode_2__get_Right(m2)
                
                elif MapTreeNode_2__get_Right(m2) is None:
                    return MapTreeNode_2__get_Left(m2)
                
                else: 
                    pattern_input : Tuple[_KEY, _VALUE, Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = MapTreeModule_spliceOutSuccessor(MapTreeNode_2__get_Right(m2))
                    return MapTreeModule_mk(MapTreeNode_2__get_Left(m2), pattern_input[0], pattern_input[1], pattern_input[2])
                
            
            else: 
                return MapTreeModule_rebalance(MapTreeNode_2__get_Left(m2), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeModule_remove(comparer, k, MapTreeNode_2__get_Right(m2)))
            
        
        elif c == 0:
            return MapTreeModule_empty()
        
        else: 
            return m
        
    
    else: 
        return MapTreeModule_empty()
    


def MapTreeModule_change(comparer: IComparer[_KEY], k: _KEY, u: Callable[[Optional[_VALUE]], Optional[_VALUE]], m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        if isinstance(m2, MapTreeNode_2):
            c : int = comparer.Compare(k, MapTreeLeaf_2__get_Key(m2)) or 0
            if c < 0:
                return MapTreeModule_rebalance(MapTreeModule_change(comparer, k, u, MapTreeNode_2__get_Left(m2)), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeNode_2__get_Right(m2))
            
            elif c == 0:
                match_value_1 : Optional[_VALUE] = u(some(MapTreeLeaf_2__get_Value(m2)))
                if match_value_1 is not None:
                    return MapTreeNode_2__ctor_Z39DE9543(k, value_1(match_value_1), MapTreeNode_2__get_Left(m2), MapTreeNode_2__get_Right(m2), MapTreeNode_2__get_Height(m2))
                
                elif MapTreeNode_2__get_Left(m2) is None:
                    return MapTreeNode_2__get_Right(m2)
                
                elif MapTreeNode_2__get_Right(m2) is None:
                    return MapTreeNode_2__get_Left(m2)
                
                else: 
                    pattern_input : Tuple[_KEY, _VALUE, Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = MapTreeModule_spliceOutSuccessor(MapTreeNode_2__get_Right(m2))
                    return MapTreeModule_mk(MapTreeNode_2__get_Left(m2), pattern_input[0], pattern_input[1], pattern_input[2])
                
            
            else: 
                return MapTreeModule_rebalance(MapTreeNode_2__get_Left(m2), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), MapTreeModule_change(comparer, k, u, MapTreeNode_2__get_Right(m2)))
            
        
        else: 
            c_1 : int = comparer.Compare(k, MapTreeLeaf_2__get_Key(m2)) or 0
            if c_1 < 0:
                match_value_2 : Optional[_VALUE] = u(None)
                if match_value_2 is not None:
                    return MapTreeNode_2__ctor_Z39DE9543(k, value_1(match_value_2), MapTreeModule_empty(), m, 2)
                
                else: 
                    return m
                
            
            elif c_1 == 0:
                match_value_3 : Optional[_VALUE] = u(some(MapTreeLeaf_2__get_Value(m2)))
                if match_value_3 is not None:
                    return MapTreeLeaf_2__ctor_5BDDA1(k, value_1(match_value_3))
                
                else: 
                    return MapTreeModule_empty()
                
            
            else: 
                match_value_4 : Optional[_VALUE] = u(None)
                if match_value_4 is not None:
                    return MapTreeNode_2__ctor_Z39DE9543(k, value_1(match_value_4), m, MapTreeModule_empty(), 2)
                
                else: 
                    return m
                
            
        
    
    else: 
        match_value : Optional[_VALUE] = u(None)
        if match_value is not None:
            return MapTreeLeaf_2__ctor_5BDDA1(k, value_1(match_value))
        
        else: 
            return m
        
    


def MapTreeModule_mem(comparer_mut: IComparer[_KEY], k_mut: _KEY, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> bool:
    while True:
        (comparer, k, m) = (comparer_mut, k_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            c : int = comparer.Compare(k, MapTreeLeaf_2__get_Key(m2)) or 0
            if isinstance(m2, MapTreeNode_2):
                if c < 0:
                    comparer_mut = comparer
                    k_mut = k
                    m_mut = MapTreeNode_2__get_Left(m2)
                    continue
                
                elif c == 0:
                    return True
                
                else: 
                    comparer_mut = comparer
                    k_mut = k
                    m_mut = MapTreeNode_2__get_Right(m2)
                    continue
                
            
            else: 
                return c == 0
            
        
        else: 
            return False
        
        break


def MapTreeModule_iterOpt(f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> None:
    while True:
        (f, m) = (f_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                MapTreeModule_iterOpt(f, MapTreeNode_2__get_Left(m2))
                f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
                f_mut = f
                m_mut = MapTreeNode_2__get_Right(m2)
                continue
            
            else: 
                f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
            
        
        break


def MapTreeModule_iter(f: Callable[[_A_, _B_], None], m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> None:
    MapTreeModule_iterOpt(f, m)


def MapTreeModule_tryPickOpt(f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> Optional[_A_]:
    while True:
        (f, m) = (f_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                match_value : Optional[_A_] = MapTreeModule_tryPickOpt(f, MapTreeNode_2__get_Left(m2))
                if match_value is None:
                    match_value_1 : Optional[_A_] = f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
                    if match_value_1 is None:
                        f_mut = f
                        m_mut = MapTreeNode_2__get_Right(m2)
                        continue
                    
                    else: 
                        return match_value_1
                    
                
                else: 
                    return match_value
                
            
            else: 
                return f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
            
        
        else: 
            return None
        
        break


def MapTreeModule_tryPick(f: Callable[[_A_, _B_], Optional[_C_]], m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> Optional[_C_]:
    return MapTreeModule_tryPickOpt(f, m)


def MapTreeModule_existsOpt(f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> bool:
    while True:
        (f, m) = (f_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                if True if MapTreeModule_existsOpt(f, MapTreeNode_2__get_Left(m2)) else f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2)):
                    return True
                
                else: 
                    f_mut = f
                    m_mut = MapTreeNode_2__get_Right(m2)
                    continue
                
            
            else: 
                return f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
            
        
        else: 
            return False
        
        break


def MapTreeModule_exists(f: Callable[[_A_, _B_], bool], m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> bool:
    return MapTreeModule_existsOpt(f, m)


def MapTreeModule_forallOpt(f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> bool:
    while True:
        (f, m) = (f_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                if f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2)) if MapTreeModule_forallOpt(f, MapTreeNode_2__get_Left(m2)) else False:
                    f_mut = f
                    m_mut = MapTreeNode_2__get_Right(m2)
                    continue
                
                else: 
                    return False
                
            
            else: 
                return f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
            
        
        else: 
            return True
        
        break


def MapTreeModule_forall(f: Callable[[_A_, _B_], bool], m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> bool:
    return MapTreeModule_forallOpt(f, m)


def MapTreeModule_map(f: Callable[[_VALUE], _RESULT], m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _RESULT]]:
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        if isinstance(m2, MapTreeNode_2):
            l2 : Optional[MapTreeLeaf_2[_KEY, _RESULT]] = MapTreeModule_map(f, MapTreeNode_2__get_Left(m2))
            v2 : _RESULT = f(MapTreeLeaf_2__get_Value(m2))
            r2 : Optional[MapTreeLeaf_2[_KEY, _RESULT]] = MapTreeModule_map(f, MapTreeNode_2__get_Right(m2))
            return MapTreeNode_2__ctor_Z39DE9543(MapTreeLeaf_2__get_Key(m2), v2, l2, r2, MapTreeNode_2__get_Height(m2))
        
        else: 
            return MapTreeLeaf_2__ctor_5BDDA1(MapTreeLeaf_2__get_Key(m2), f(MapTreeLeaf_2__get_Value(m2)))
        
    
    else: 
        return MapTreeModule_empty()
    


def MapTreeModule_mapiOpt(f: Any, m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> Optional[MapTreeLeaf_2[_KEY, _RESULT]]:
    if m is not None:
        m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
        if isinstance(m2, MapTreeNode_2):
            l2 : Optional[MapTreeLeaf_2[_KEY, _RESULT]] = MapTreeModule_mapiOpt(f, MapTreeNode_2__get_Left(m2))
            v2 : _RESULT = f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
            r2 : Optional[MapTreeLeaf_2[_KEY, _RESULT]] = MapTreeModule_mapiOpt(f, MapTreeNode_2__get_Right(m2))
            return MapTreeNode_2__ctor_Z39DE9543(MapTreeLeaf_2__get_Key(m2), v2, l2, r2, MapTreeNode_2__get_Height(m2))
        
        else: 
            return MapTreeLeaf_2__ctor_5BDDA1(MapTreeLeaf_2__get_Key(m2), f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2)))
        
    
    else: 
        return MapTreeModule_empty()
    


def MapTreeModule_mapi(f: Callable[[_A_, _B_], _C_], m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> Optional[MapTreeLeaf_2[_A_, _C_]]:
    return MapTreeModule_mapiOpt(f, m)


def MapTreeModule_foldBackOpt(f_mut: Any, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]], x_mut: _A_) -> _A_:
    while True:
        (f, m, x) = (f_mut, m_mut, x_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                x_1 : _A_ = MapTreeModule_foldBackOpt(f, MapTreeNode_2__get_Right(m2), x)
                x_2 : _A_ = f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), x_1)
                f_mut = f
                m_mut = MapTreeNode_2__get_Left(m2)
                x_mut = x_2
                continue
            
            else: 
                return f(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), x)
            
        
        else: 
            return x
        
        break


def MapTreeModule_foldBack(f: Callable[[_A_, _B_, _C_], _C_], m: Optional[MapTreeLeaf_2[_A_, _B_]], x: _C_) -> _C_:
    return MapTreeModule_foldBackOpt(f, m, x)


def MapTreeModule_foldOpt(f_mut: Any, x_mut: _A_, m_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]]) -> _A_:
    while True:
        (f, x, m) = (f_mut, x_mut, m_mut)
        if m is not None:
            m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
            if isinstance(m2, MapTreeNode_2):
                f_mut = f
                x_mut = f(MapTreeModule_foldOpt(f, x, MapTreeNode_2__get_Left(m2)), MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
                m_mut = MapTreeNode_2__get_Right(m2)
                continue
            
            else: 
                return f(x, MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2))
            
        
        else: 
            return x
        
        break


def MapTreeModule_fold(f: Callable[[_A_, _B_, _C_], _A_], x: _A_, m: Optional[MapTreeLeaf_2[_B_, _C_]]=None) -> _A_:
    return MapTreeModule_foldOpt(f, x, m)


def MapTreeModule_foldSectionOpt(comparer: IComparer[_KEY], lo: _KEY, hi: _KEY, f: Any, m: Optional[MapTreeLeaf_2[_KEY, _VALUE]], x: _A_) -> _A_:
    def fold_from_to(f_1_mut: Any, m_1_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]], x_1_mut: _B_, comparer: IComparer[_KEY]=comparer, lo: _KEY=lo, hi: _KEY=hi, f: Any=f, m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=m, x: _A_=x) -> _B_:
        while True:
            (f_1, m_1, x_1) = (f_1_mut, m_1_mut, x_1_mut)
            if m_1 is not None:
                m2 : MapTreeLeaf_2[_KEY, _VALUE] = m_1
                if isinstance(m2, MapTreeNode_2):
                    c_lo_key : int = comparer.Compare(lo, MapTreeLeaf_2__get_Key(m2)) or 0
                    c_key_hi : int = comparer.Compare(MapTreeLeaf_2__get_Key(m2), hi) or 0
                    x_2 : _B_ = fold_from_to(f_1, MapTreeNode_2__get_Left(m2), x_1) if (c_lo_key < 0) else x_1
                    x_3 : _B_ = f_1(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), x_2) if ((c_key_hi <= 0) if (c_lo_key <= 0) else False) else x_2
                    if c_key_hi < 0:
                        f_1_mut = f_1
                        m_1_mut = MapTreeNode_2__get_Right(m2)
                        x_1_mut = x_3
                        continue
                    
                    else: 
                        return x_3
                    
                
                elif (comparer.Compare(MapTreeLeaf_2__get_Key(m2), hi) <= 0) if (comparer.Compare(lo, MapTreeLeaf_2__get_Key(m2)) <= 0) else False:
                    return f_1(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2), x_1)
                
                else: 
                    return x_1
                
            
            else: 
                return x_1
            
            break
    
    if comparer.Compare(lo, hi) == 1:
        return x
    
    else: 
        return fold_from_to(f, m, x)
    


def MapTreeModule_foldSection(comparer: IComparer[_KEY], lo: _KEY, hi: _KEY, f: Callable[[_KEY, _A_, _B_], _B_], m: Optional[MapTreeLeaf_2[_KEY, _A_]], x: _B_) -> _B_:
    return MapTreeModule_foldSectionOpt(comparer, lo, hi, f, m, x)


def MapTreeModule_toList(m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> FSharpList[Tuple[_KEY, _VALUE]]:
    def loop(m_1_mut: Optional[MapTreeLeaf_2[_KEY, _VALUE]], acc_mut: FSharpList[Tuple[_KEY, _VALUE]], m: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=m) -> FSharpList[Tuple[_KEY, _VALUE]]:
        while True:
            (m_1, acc) = (m_1_mut, acc_mut)
            if m_1 is not None:
                m2 : MapTreeLeaf_2[_KEY, _VALUE] = m_1
                if isinstance(m2, MapTreeNode_2):
                    m_1_mut = MapTreeNode_2__get_Left(m2)
                    acc_mut = cons((MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2)), loop(MapTreeNode_2__get_Right(m2), acc))
                    continue
                
                else: 
                    return cons((MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2)), acc)
                
            
            else: 
                return acc
            
            break
    
    return loop(m, empty_1())


def MapTreeModule_copyToArray(m: Optional[MapTreeLeaf_2[_A_, _B_]], arr: List[Any], i: int) -> None:
    j : int = i or 0
    def arrow_270(x: _A_, y: _B_, m: Optional[MapTreeLeaf_2[_A_, _B_]]=m, arr: List[Any]=arr, i: int=i) -> None:
        nonlocal j
        arr[j] = (x, y)
        j = (j + 1) or 0
    
    MapTreeModule_iter(arrow_270, m)


def MapTreeModule_toArray(m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> List[Any]:
    n : int = MapTreeModule_size(m) or 0
    res : List[Any] = fill([0] * n, 0, n, (None, None))
    MapTreeModule_copyToArray(m, res, 0)
    return res


def MapTreeModule_ofList(comparer: IComparer[_A_], l: FSharpList[Tuple[_A_, _B_]]) -> Optional[MapTreeLeaf_2[_A_, _B_]]:
    def arrow_271(acc: Optional[MapTreeLeaf_2[_A_, _B_]], tupled_arg: Tuple[_A_, _B_], comparer: IComparer[_A_]=comparer, l: FSharpList[Tuple[_A_, _B_]]=l) -> Optional[MapTreeLeaf_2[_A_, _B_]]:
        return MapTreeModule_add(comparer, tupled_arg[0], tupled_arg[1], acc)
    
    return fold_1(arrow_271, MapTreeModule_empty(), l)


def MapTreeModule_mkFromEnumerator(comparer_mut: IComparer[_A_], acc_mut: Optional[MapTreeLeaf_2[_A_, _B_]], e_mut: IEnumerator[Tuple[_A_, _B_]]) -> Optional[MapTreeLeaf_2[_A_, _B_]]:
    while True:
        (comparer, acc, e) = (comparer_mut, acc_mut, e_mut)
        if e.System_Collections_IEnumerator_MoveNext():
            pattern_input : Tuple[_A_, _B_] = e.System_Collections_Generic_IEnumerator_00601_get_Current()
            comparer_mut = comparer
            acc_mut = MapTreeModule_add(comparer, pattern_input[0], pattern_input[1], acc)
            e_mut = e
            continue
        
        else: 
            return acc
        
        break


def MapTreeModule_ofArray(comparer: IComparer[_KEY], arr: List[Tuple[_KEY, _VALUE]]) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    res : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = MapTreeModule_empty()
    for idx in range(0, (len(arr) - 1) + 1, 1):
        for_loop_var : Tuple[_KEY, _VALUE] = arr[idx]
        res = MapTreeModule_add(comparer, for_loop_var[0], for_loop_var[1], res)
    return res


def MapTreeModule_ofSeq(comparer: IComparer[_KEY], c: IEnumerable[Tuple[_KEY, _VALUE]]) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    if is_array_like(c):
        return MapTreeModule_ofArray(comparer, c)
    
    elif isinstance(c, FSharpList):
        return MapTreeModule_ofList(comparer, c)
    
    else: 
        with get_enumerator(c) as ie:
            return MapTreeModule_mkFromEnumerator(comparer, MapTreeModule_empty(), ie)
    


def expr_272(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return record_type("Map.MapTreeModule.MapIterator`2", [gen0, gen1], MapTreeModule_MapIterator_2, lambda: [("stack", list_type(option_type(MapTreeLeaf_2_reflection(gen0, gen1)))), ("started", bool_type)])


class MapTreeModule_MapIterator_2(Record, Generic[_KEY, _VALUE]):
    def __init__(self, stack: FSharpList[Optional[MapTreeLeaf_2[_KEY, _VALUE]]], started: bool) -> None:
        super().__init__()
        self.stack = stack
        self.started = started
    

MapTreeModule_MapIterator_2_reflection = expr_272

def MapTreeModule_collapseLHS(stack_mut: FSharpList[Optional[MapTreeLeaf_2[_KEY, _VALUE]]]) -> FSharpList[Optional[MapTreeLeaf_2[_KEY, _VALUE]]]:
    while True:
        (stack,) = (stack_mut,)
        if not is_empty_1(stack):
            rest : FSharpList[Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = tail(stack)
            m : Optional[MapTreeLeaf_2[_KEY, _VALUE]] = head(stack)
            if m is not None:
                m2 : MapTreeLeaf_2[_KEY, _VALUE] = m
                if isinstance(m2, MapTreeNode_2):
                    stack_mut = of_array_with_tail([MapTreeNode_2__get_Left(m2), MapTreeLeaf_2__ctor_5BDDA1(MapTreeLeaf_2__get_Key(m2), MapTreeLeaf_2__get_Value(m2)), MapTreeNode_2__get_Right(m2)], rest)
                    continue
                
                else: 
                    return stack
                
            
            else: 
                stack_mut = rest
                continue
            
        
        else: 
            return empty_1()
        
        break


def MapTreeModule_mkIterator(m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> MapTreeModule_MapIterator_2[_A_, _B_]:
    return MapTreeModule_MapIterator_2(MapTreeModule_collapseLHS(singleton(m)), False)


def MapTreeModule_notStarted() -> _A_:
    raise Exception("enumeration not started")


def MapTreeModule_alreadyFinished() -> _A_:
    raise Exception("enumeration already finished")


def MapTreeModule_current(i: MapTreeModule_MapIterator_2[_KEY, _VALUE]) -> Any:
    if i.started:
        match_value : FSharpList[Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = i.stack
        if not is_empty_1(match_value):
            if head(match_value) is not None:
                m : MapTreeLeaf_2[_KEY, _VALUE] = head(match_value)
                if isinstance(m, MapTreeNode_2):
                    raise Exception("Please report error: Map iterator, unexpected stack for current")
                
                else: 
                    return (MapTreeLeaf_2__get_Key(m), MapTreeLeaf_2__get_Value(m))
                
            
            else: 
                raise Exception("Please report error: Map iterator, unexpected stack for current")
            
        
        else: 
            return MapTreeModule_alreadyFinished()
        
    
    else: 
        return MapTreeModule_notStarted()
    


def MapTreeModule_moveNext(i: MapTreeModule_MapIterator_2[_KEY, _VALUE]) -> bool:
    if i.started:
        match_value : FSharpList[Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = i.stack
        if not is_empty_1(match_value):
            if head(match_value) is not None:
                m : MapTreeLeaf_2[_KEY, _VALUE] = head(match_value)
                if isinstance(m, MapTreeNode_2):
                    raise Exception("Please report error: Map iterator, unexpected stack for moveNext")
                
                else: 
                    i.stack = MapTreeModule_collapseLHS(tail(match_value))
                    return not is_empty_1(i.stack)
                
            
            else: 
                raise Exception("Please report error: Map iterator, unexpected stack for moveNext")
            
        
        else: 
            return False
        
    
    else: 
        i.started = True
        return not is_empty_1(i.stack)
    


def MapTreeModule_mkIEnumerator(m: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> IEnumerator[Any]:
    i : MapTreeModule_MapIterator_2[_A_, _B_] = MapTreeModule_mkIterator(m)
    class ObjectExpr273(IEnumerator[Any]):
        def System_Collections_Generic_IEnumerator_00601_get_Current(self, m: Optional[MapTreeLeaf_2[_A_, _B_]]=m) -> Any:
            return MapTreeModule_current(i)
        
        def System_Collections_IEnumerator_get_Current(self, m: Optional[MapTreeLeaf_2[_A_, _B_]]=m) -> Any:
            return MapTreeModule_current(i)
        
        def System_Collections_IEnumerator_MoveNext(self, m: Optional[MapTreeLeaf_2[_A_, _B_]]=m) -> bool:
            return MapTreeModule_moveNext(i)
        
        def System_Collections_IEnumerator_Reset(self, m: Optional[MapTreeLeaf_2[_A_, _B_]]=m) -> None:
            nonlocal i
            i = MapTreeModule_mkIterator(m)
        
        def Dispose(self, m: Optional[MapTreeLeaf_2[_A_, _B_]]=m) -> None:
            pass
        
    return ObjectExpr273()


def MapTreeModule_toSeq(s: Optional[MapTreeLeaf_2[_A_, _B_]]=None) -> IEnumerable[Any]:
    def generator(en_1: IEnumerator[Any], s: Optional[MapTreeLeaf_2[_A_, _B_]]=s) -> Optional[Tuple[Any, IEnumerator[Any]]]:
        if en_1.System_Collections_IEnumerator_MoveNext():
            return (en_1.System_Collections_Generic_IEnumerator_00601_get_Current(), en_1)
        
        else: 
            return None
        
    
    return unfold(generator, MapTreeModule_mkIEnumerator(s))


def expr_276(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return class_type("Map.FSharpMap", [gen0, gen1], FSharpMap)


class FSharpMap(Generic[_KEY, _VALUE]):
    def __init__(self, comparer: IComparer[_KEY], tree: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> None:
        self.comparer = comparer
        self.tree = tree
    
    def GetHashCode(self) -> int:
        this : FSharpMap[_KEY, _VALUE] = self
        return FSharpMap__ComputeHashCode(this)
    
    def __eq__(self, that: Any=None) -> bool:
        this : FSharpMap[_KEY, _VALUE] = self
        if isinstance(that, FSharpMap):
            with get_enumerator(this) as e1:
                with get_enumerator(that) as e2:
                    def loop(__unit: Any=None) -> bool:
                        m1 : bool = e1.System_Collections_IEnumerator_MoveNext()
                        if m1 == e2.System_Collections_IEnumerator_MoveNext():
                            if not m1:
                                return True
                            
                            else: 
                                e1c : Any = e1.System_Collections_Generic_IEnumerator_00601_get_Current()
                                e2c : Any = e2.System_Collections_Generic_IEnumerator_00601_get_Current()
                                if equals(e1c[1], e2c[1]) if equals(e1c[0], e2c[0]) else False:
                                    return loop()
                                
                                else: 
                                    return False
                                
                            
                        
                        else: 
                            return False
                        
                    
                    loop : Callable[[], bool] = loop
                    return loop()
        
        else: 
            return False
        
    
    def __str__(self) -> str:
        this : FSharpMap[_KEY, _VALUE] = self
        def arrow_274(kv: Any) -> str:
            return format("({0}, {1})", kv[0], kv[1])
        
        return ("map [" + join("; ", map_1(arrow_274, this))) + "]"
    
    @property
    def Symbol_toStringTag(self) -> str:
        return "FSharpMap"
    
    def to_json(self, _key: str) -> Any:
        this : FSharpMap[_KEY, _VALUE] = self
        return Array.from_(this)
    
    def GetEnumerator(self) -> IEnumerator[Any]:
        _ : FSharpMap[_KEY, _VALUE] = self
        return MapTreeModule_mkIEnumerator(_.tree)
    
    def __iter__(self) -> IEnumerator[Any]:
        return to_iterator(self.GetEnumerator())
    
    def System_Collections_IEnumerable_GetEnumerator(self) -> IEnumerator[Any]:
        _ : FSharpMap[_KEY, _VALUE] = self
        return MapTreeModule_mkIEnumerator(_.tree)
    
    def CompareTo(self, obj: Any=None) -> int:
        m : FSharpMap[_KEY, _VALUE] = self
        if isinstance(obj, FSharpMap):
            def arrow_275(kvp1: Any, kvp2: Any) -> int:
                c : int = m.comparer.Compare(kvp1[0], kvp2[0]) or 0
                return c if (c != 0) else compare(kvp1[1], kvp2[1])
            
            return compare_with(arrow_275, m, obj)
        
        else: 
            raise Exception("not comparable\\nParameter name: obj")
        
    
    def System_Collections_Generic_ICollection_00601_Add2B595(self, x: Any) -> None:
        ignore(x)
        raise Exception("Map cannot be mutated")
    
    def System_Collections_Generic_ICollection_00601_Clear(self) -> None:
        raise Exception("Map cannot be mutated")
    
    def System_Collections_Generic_ICollection_00601_Remove2B595(self, x: Any) -> bool:
        ignore(x)
        raise Exception("Map cannot be mutated")
    
    def System_Collections_Generic_ICollection_00601_Contains2B595(self, x: Any) -> bool:
        m : FSharpMap[_KEY, _VALUE] = self
        return equals(FSharpMap__get_Item(m, x[0]), x[1]) if FSharpMap__ContainsKey(m, x[0]) else False
    
    def System_Collections_Generic_ICollection_00601_CopyToZ2E171D71(self, arr: List[Any], i: int) -> None:
        m : FSharpMap[_KEY, _VALUE] = self
        MapTreeModule_copyToArray(m.tree, arr, i)
    
    def System_Collections_Generic_ICollection_00601_get_IsReadOnly(self) -> bool:
        return True
    
    def System_Collections_Generic_ICollection_00601_get_Count(self) -> int:
        m : FSharpMap[_KEY, _VALUE] = self
        return FSharpMap__get_Count(m)
    
    def System_Collections_Generic_IReadOnlyCollection_00601_get_Count(self) -> int:
        m : FSharpMap[_KEY, _VALUE] = self
        return FSharpMap__get_Count(m)
    
    @property
    def size(self) -> int:
        m : FSharpMap[_KEY, _VALUE] = self
        return FSharpMap__get_Count(m)
    
    def clear(self) -> None:
        raise Exception("Map cannot be mutated")
    
    def delete(self, _arg1: Optional[_KEY]=None) -> bool:
        raise Exception("Map cannot be mutated")
        return False
    
    def entries(self) -> IEnumerable[Tuple[_KEY, _VALUE]]:
        m : FSharpMap[_KEY, _VALUE] = self
        def mapping(p: Any) -> Tuple[_KEY, _VALUE]:
            return (p[0], p[1])
        
        return map_1(mapping, m)
    
    def __getitem__(self, k: Optional[_KEY]=None) -> _VALUE:
        m : FSharpMap[_KEY, _VALUE] = self
        return FSharpMap__get_Item(m, k)
    
    def has(self, k: Optional[_KEY]=None) -> bool:
        m : FSharpMap[_KEY, _VALUE] = self
        return FSharpMap__ContainsKey(m, k)
    
    def keys(self) -> IEnumerable[_KEY]:
        m : FSharpMap[_KEY, _VALUE] = self
        def mapping(p: Any) -> _KEY:
            return p[0]
        
        return map_1(mapping, m)
    
    def __setitem__(self, k: _KEY, v: _VALUE) -> Map_2[_KEY, _VALUE]:
        m : FSharpMap[_KEY, _VALUE] = self
        raise Exception("Map cannot be mutated")
        return m
    
    def values(self) -> IEnumerable[_VALUE]:
        m : FSharpMap[_KEY, _VALUE] = self
        def mapping(p: Any) -> _VALUE:
            return p[1]
        
        return map_1(mapping, m)
    
    def for_each(self, f: Callable[[_VALUE, _KEY, Map_2[_KEY, _VALUE]], None], this_arg: Optional[Any]=None) -> None:
        m : FSharpMap[_KEY, _VALUE] = self
        def action(p: Any) -> None:
            f(p[1], p[0], m)
        
        iterate_1(action, m)
    

FSharpMap_reflection = expr_276

def FSharpMap__ctor(comparer: IComparer[_KEY], tree: Optional[MapTreeLeaf_2[_KEY, _VALUE]]=None) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap(comparer, tree)


def FSharpMap_Empty(comparer: IComparer[_KEY]) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(comparer, MapTreeModule_empty())


def FSharpMap__get_Comparer(m: FSharpMap[_KEY, _VALUE]) -> IComparer[_KEY]:
    return m.comparer


def FSharpMap__get_Tree(m: FSharpMap[_KEY, _VALUE]) -> Optional[MapTreeLeaf_2[_KEY, _VALUE]]:
    return m.tree


def FSharpMap__Add(m: FSharpMap[_KEY, _VALUE], key: _KEY, value: _VALUE) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(m.comparer, MapTreeModule_add(m.comparer, key, value, m.tree))


def FSharpMap__Change(m: FSharpMap[_KEY, _VALUE], key: _KEY, f: Callable[[Optional[_VALUE]], Optional[_VALUE]]) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(m.comparer, MapTreeModule_change(m.comparer, key, f, m.tree))


def FSharpMap__get_IsEmpty(m: FSharpMap[_KEY, _VALUE]) -> bool:
    return m.tree is None


def FSharpMap__get_Item(m: FSharpMap[_KEY, _VALUE], key: _KEY) -> _VALUE:
    return MapTreeModule_find(m.comparer, key, m.tree)


def FSharpMap__TryPick(m: FSharpMap[_KEY, _VALUE], f: Callable[[_KEY, _VALUE], Optional[_A_]]) -> Optional[_A_]:
    return MapTreeModule_tryPick(f, m.tree)


def FSharpMap__Exists(m: FSharpMap[_KEY, _VALUE], predicate: Callable[[_KEY, _VALUE], bool]) -> bool:
    return MapTreeModule_exists(predicate, m.tree)


def FSharpMap__Filter(m: FSharpMap[_KEY, _VALUE], predicate: Callable[[_KEY, _VALUE], bool]) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(m.comparer, MapTreeModule_filter(m.comparer, predicate, m.tree))


def FSharpMap__ForAll(m: FSharpMap[_KEY, _VALUE], predicate: Callable[[_KEY, _VALUE], bool]) -> bool:
    return MapTreeModule_forall(predicate, m.tree)


def FSharpMap__Fold(m: FSharpMap[_KEY, _VALUE], f: Callable[[_KEY, _VALUE, _A_], _A_], acc: _A_) -> _A_:
    return MapTreeModule_foldBack(f, m.tree, acc)


def FSharpMap__FoldSection(m: FSharpMap[_KEY, _VALUE], lo: _KEY, hi: _KEY, f: Callable[[_KEY, _VALUE, _Z], _Z], acc: _Z) -> _Z:
    return MapTreeModule_foldSection(m.comparer, lo, hi, f, m.tree, acc)


def FSharpMap__Iterate(m: FSharpMap[_KEY, _VALUE], f: Callable[[_KEY, _VALUE], None]) -> None:
    MapTreeModule_iter(f, m.tree)


def FSharpMap__MapRange(m: FSharpMap[_KEY, _VALUE], f: Callable[[_VALUE], _RESULT]) -> FSharpMap[_KEY, _RESULT]:
    return FSharpMap__ctor(m.comparer, MapTreeModule_map(f, m.tree))


def FSharpMap__Map(m: FSharpMap[_KEY, _VALUE], f: Callable[[_KEY, _VALUE], _B]) -> FSharpMap[_KEY, _B]:
    return FSharpMap__ctor(m.comparer, MapTreeModule_mapi(f, m.tree))


def FSharpMap__Partition(m: FSharpMap[_KEY, _VALUE], predicate: Callable[[_KEY, _VALUE], bool]) -> Tuple[FSharpMap[_KEY, _VALUE], FSharpMap[_KEY, _VALUE]]:
    pattern_input : Tuple[Optional[MapTreeLeaf_2[_KEY, _VALUE]], Optional[MapTreeLeaf_2[_KEY, _VALUE]]] = MapTreeModule_partition(m.comparer, predicate, m.tree)
    return (FSharpMap__ctor(m.comparer, pattern_input[0]), FSharpMap__ctor(m.comparer, pattern_input[1]))


def FSharpMap__get_Count(m: FSharpMap[_KEY, _VALUE]) -> int:
    return MapTreeModule_size(m.tree)


def FSharpMap__ContainsKey(m: FSharpMap[_KEY, _VALUE], key: _KEY) -> bool:
    return MapTreeModule_mem(m.comparer, key, m.tree)


def FSharpMap__Remove(m: FSharpMap[_KEY, _VALUE], key: _KEY) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(m.comparer, MapTreeModule_remove(m.comparer, key, m.tree))


def FSharpMap__TryGetValue(_: FSharpMap[_KEY, _VALUE], key: _KEY, value: FSharpRef[_VALUE]) -> bool:
    match_value : Optional[_VALUE] = MapTreeModule_tryFind(_.comparer, key, _.tree)
    if match_value is None:
        return False
    
    else: 
        v : _VALUE = value_1(match_value)
        value.contents = v
        return True
    


def FSharpMap__get_Keys(__: FSharpMap[_KEY, _VALUE]) -> ICollection[_KEY]:
    def mapping(kvp: Any, __: FSharpMap[_KEY, _VALUE]=__) -> _KEY:
        return kvp[0]
    
    return map_2(mapping, MapTreeModule_toArray(__.tree), None)


def FSharpMap__get_Values(__: FSharpMap[_KEY, _VALUE]) -> ICollection[_VALUE]:
    def mapping(kvp: Any, __: FSharpMap[_KEY, _VALUE]=__) -> _VALUE:
        return kvp[1]
    
    return map_2(mapping, MapTreeModule_toArray(__.tree), None)


def FSharpMap__TryFind(m: FSharpMap[_KEY, _VALUE], key: _KEY) -> Optional[_VALUE]:
    return MapTreeModule_tryFind(m.comparer, key, m.tree)


def FSharpMap__ToList(m: FSharpMap[_KEY, _VALUE]) -> FSharpList[Tuple[_KEY, _VALUE]]:
    return MapTreeModule_toList(m.tree)


def FSharpMap__ToArray(m: FSharpMap[_KEY, _VALUE]) -> List[Any]:
    return MapTreeModule_toArray(m.tree)


def FSharpMap__ComputeHashCode(this: FSharpMap[_KEY, _VALUE]) -> int:
    def combine_hash(x: int, y: int, this: FSharpMap[_KEY, _VALUE]=this) -> int:
        return ((x << 1) + y) + 631
    
    res : int = 0
    with get_enumerator(this) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            active_pattern_result6638 : Tuple[_KEY, _VALUE] = enumerator.System_Collections_Generic_IEnumerator_00601_get_Current()
            res = combine_hash(res, structural_hash(active_pattern_result6638[0])) or 0
            res = combine_hash(res, structural_hash(active_pattern_result6638[1])) or 0
    return res


def is_empty(table: FSharpMap[_A_, _B_]) -> bool:
    return FSharpMap__get_IsEmpty(table)


def add(key: _A_, value: _B_, table: FSharpMap[_A_, _B_]) -> FSharpMap[_A_, _B_]:
    return FSharpMap__Add(table, key, value)


def change(key: _A_, f: Callable[[Optional[_B_]], Optional[_B_]], table: FSharpMap[_A_, _B_]) -> FSharpMap[_A_, _B_]:
    return FSharpMap__Change(table, key, f)


def find(key: _A_, table: FSharpMap[_A_, _B_]) -> _B_:
    return FSharpMap__get_Item(table, key)


def try_find(key: _A_, table: FSharpMap[_A_, _B_]) -> Optional[_B_]:
    return FSharpMap__TryFind(table, key)


def remove(key: _A_, table: FSharpMap[_A_, _B_]) -> FSharpMap[_A_, _B_]:
    return FSharpMap__Remove(table, key)


def contains_key(key: _A_, table: FSharpMap[_A_, _B_]) -> bool:
    return FSharpMap__ContainsKey(table, key)


def iterate(action: Callable[[_A_, _B_], None], table: FSharpMap[_A_, _B_]) -> None:
    FSharpMap__Iterate(table, action)


def try_pick(chooser: Callable[[_A_, _B_], Optional[_C_]], table: FSharpMap[_A_, _B_]) -> Optional[_C_]:
    return FSharpMap__TryPick(table, chooser)


def pick(chooser: Callable[[_A_, _B_], Optional[_C_]], table: FSharpMap[_A_, _B_]) -> _C_:
    match_value : Optional[_C_] = try_pick(chooser, table)
    if match_value is not None:
        return value_1(match_value)
    
    else: 
        raise Exception()
    


def exists(predicate: Callable[[_A_, _B_], bool], table: FSharpMap[_A_, _B_]) -> bool:
    return FSharpMap__Exists(table, predicate)


def filter(predicate: Callable[[_A_, _B_], bool], table: FSharpMap[_A_, _B_]) -> FSharpMap[_A_, _B_]:
    return FSharpMap__Filter(table, predicate)


def partition(predicate: Callable[[_A_, _B_], bool], table: FSharpMap[_A_, _B_]) -> Tuple[FSharpMap[_A_, _B_], FSharpMap[_A_, _B_]]:
    return FSharpMap__Partition(table, predicate)


def for_all(predicate: Callable[[_A_, _B_], bool], table: FSharpMap[_A_, _B_]) -> bool:
    return FSharpMap__ForAll(table, predicate)


def map(mapping: Callable[[_A_, _B_], _C_], table: FSharpMap[_A_, _B_]) -> FSharpMap[_A_, _C_]:
    return FSharpMap__Map(table, mapping)


def fold(folder: Callable[[_STATE, _KEY, _T], _STATE], state: _STATE, table: FSharpMap[_KEY, _T]) -> _STATE:
    return MapTreeModule_fold(folder, state, FSharpMap__get_Tree(table))


def fold_back(folder: Callable[[_KEY, _T, _STATE], _STATE], table: FSharpMap[_KEY, _T], state: _STATE) -> _STATE:
    return MapTreeModule_foldBack(folder, FSharpMap__get_Tree(table), state)


def to_seq(table: FSharpMap[_A_, _B_]) -> IEnumerable[Tuple[_A_, _B_]]:
    def mapping(kvp: Any, table: FSharpMap[_A_, _B_]=table) -> Tuple[_A_, _B_]:
        return (kvp[0], kvp[1])
    
    return map_1(mapping, table)


def find_key(predicate: Callable[[_A_, _B_], bool], table: FSharpMap[_A_, _B_]) -> _A_:
    def chooser(kvp: Any, predicate: Callable[[_A_, _B_], bool]=predicate, table: FSharpMap[_A_, _B_]=table) -> Optional[_A_]:
        k : _A_ = kvp[0]
        if predicate(k, kvp[1]):
            return some(k)
        
        else: 
            return None
        
    
    return pick_1(chooser, table)


def try_find_key(predicate: Callable[[_A_, _B_], bool], table: FSharpMap[_A_, _B_]) -> Optional[_A_]:
    def chooser(kvp: Any, predicate: Callable[[_A_, _B_], bool]=predicate, table: FSharpMap[_A_, _B_]=table) -> Optional[_A_]:
        k : _A_ = kvp[0]
        if predicate(k, kvp[1]):
            return some(k)
        
        else: 
            return None
        
    
    return try_pick_1(chooser, table)


def of_list(elements: FSharpList[Tuple[_KEY, _VALUE]], comparer: IComparer[_KEY]) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(comparer, MapTreeModule_ofSeq(comparer, elements))


def of_seq(elements: IEnumerable[Tuple[_T, _A_]], comparer: IComparer[_T]) -> FSharpMap[_T, _A_]:
    return FSharpMap__ctor(comparer, MapTreeModule_ofSeq(comparer, elements))


def of_array(elements: List[Tuple[_KEY, _VALUE]], comparer: IComparer[_KEY]) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap__ctor(comparer, MapTreeModule_ofSeq(comparer, elements))


def to_list(table: FSharpMap[_A_, _B_]) -> FSharpList[Tuple[_A_, _B_]]:
    return FSharpMap__ToList(table)


def to_array(table: FSharpMap[_A_, _B_]) -> List[Any]:
    return FSharpMap__ToArray(table)


def keys(table: FSharpMap[_K, _V]) -> ICollection[_K]:
    return FSharpMap__get_Keys(table)


def values(table: FSharpMap[_K, _V]) -> ICollection[_V]:
    return FSharpMap__get_Values(table)


def empty(comparer: IComparer[_KEY]) -> FSharpMap[_KEY, _VALUE]:
    return FSharpMap_Empty(comparer)


def count(table: FSharpMap[_A_, _B_]) -> int:
    return FSharpMap__get_Count(table)





import asyncio
from concurrent.futures import ThreadPoolExecutor
from asyncio import Future, ensure_future
from threading import Timer
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

from .async_builder import (
    Async,
    CancellationToken,
    IAsyncContext,
    OperationCanceledError,
    Trampoline,
    empty_continuation,
    protected_bind,
    protected_cont,
    protected_return,
)

# F# generated code (from Choice.fs)
from .choice import Choice_makeChoice1Of2, Choice_makeChoice2Of2  # type: ignore
from .task import TaskCompletionSource

_T = TypeVar("_T")


default_cancellation_token = CancellationToken()

# see AsyncBuilder.Delay
def delay(generator: Callable[[], Async[_T]]):
    def cont(ctx: IAsyncContext[_T]):
        generator()(ctx)

    return protected_cont(cont)


def create_cancellation_token(arg: Union[int, bool, None] = None) -> CancellationToken:
    cancelled = arg if isinstance(arg, bool) else False
    token = CancellationToken(cancelled)
    if isinstance(arg, int):
        timer = Timer(arg / 1000.0, token.cancel)  # type: ignore
        timer.start()

    return token


def cancel(token: CancellationToken) -> None:
    token.cancel()


def cancel_after(token: CancellationToken, ms: int) -> None:
    timer = Timer(ms / 1000.0, token.cancel)
    timer.start()


def is_cancellation_requested(token: CancellationToken) -> bool:
    return token and token.is_cancelled


def sleep(millisecondsDueTime: int) -> Async[_T]:
    def cont(ctx: IAsyncContext[_T]):
        def cancel():
            timer.cancel()
            ctx.on_cancel(OperationCanceledError())

        token_id = ctx.cancel_token.add_listener(cancel)

        def timeout():
            ctx.cancel_token.remove_listener(token_id)
            ctx.on_success()

        timer = Timer(millisecondsDueTime / 1000.0, timeout)
        timer.start()

    return protected_cont(cont)


def ignore(computation: Async[_T]) -> Async[None]:
    def binder(_: Optional[_T] = None) -> Async[None]:
        return protected_return()

    return protected_bind(computation, binder)


def parallel(computations: Iterable[Async[_T]]) -> Async[List[_T]]:
    def delayed() -> Async[List[_T]]:
        if TYPE_CHECKING:
            all: Future[List[_T]] = asyncio.gather(*map(start_as_task, computations))
        else:
            all: Future = asyncio.gather(*map(start_as_task, computations))

        return await_task(all)

    return delay(delayed)


def sequential(computations: Iterable[Async[_T]]) -> Async[List[_T]]:
    def delayed() -> Async[List[_T]]:
        async def sequence() -> List[_T]:
            results: List[_T] = []

            for c in computations:
                result = await start_as_task(c)
                results.append(result)

            return results

        return await_task(sequence())

    return delay(delayed)


def catch_async(work: Async[_T]) -> Async[_T]:
    def cont(ctx: IAsyncContext[_T]) -> None:
        def on_success(x: Optional[_T] = None):
            ctx.on_success(Choice_makeChoice1Of2(x))  # type: ignore

        def on_error(err: Exception):
            ctx.on_success(Choice_makeChoice2Of2(err))  # type: ignore

        ctx_ = IAsyncContext.create(
            on_success, on_error, ctx.on_cancel, ctx.trampoline, ctx.cancel_token
        )
        work(ctx_)

    return protected_cont(cont)


def from_continuations(
    f: Callable[[List[Callable[[Any], None]]], None]
) -> Callable[[IAsyncContext[Any]], None]:
    def cont(ctx: IAsyncContext[Any]) -> None:
        f([ctx.on_success, ctx.on_error, ctx.on_cancel])

    return protected_cont(cont)


def await_task(task: Awaitable[_T]) -> Async[_T]:
    """Return an asynchronous computation that will wait for the given
    task to complete and return its result.
    """
    continuation: List[Callable[[Any], None]] = []
    task = ensure_future(task)

    if TYPE_CHECKING:

        def done(tsk: Future[_T]) -> None:
            try:
                value = tsk.result()
            except Exception as ex:
                continuation[1](ex)
            else:
                continuation[0](value)

    else:

        def done(tsk: Future) -> None:
            try:
                value = tsk.result()
            except Exception as ex:
                continuation[1](ex)
            else:
                continuation[0](value)

    def callback(conts: List[Callable[[Any], None]]) -> None:
        nonlocal continuation
        continuation = conts

    task.add_done_callback(done)
    return from_continuations(callback)


def start_with_continuations(
    computation: Callable[[IAsyncContext[_T]], None],
    continuation: Optional[Callable[[Optional[_T]], None]] = None,
    exception_continuation: Optional[Callable[[Exception], None]] = None,
    cancellation_continuation: Optional[
        Callable[[OperationCanceledError], None]
    ] = None,
    cancellation_token: Optional[CancellationToken] = None,
) -> None:
    """Runs an asynchronous computation.

    Runs an asynchronous computation, starting immediately on the
    current operating system thread. Call one of the three continuations
    when the operation completes.

    If no cancellation token is provided then the default cancellation
    token is used."""
    trampoline = Trampoline()

    ctx = IAsyncContext.create(
        continuation or empty_continuation,
        exception_continuation or empty_continuation,
        cancellation_continuation or empty_continuation,
        trampoline,
        cancellation_token or default_cancellation_token,
    )

    return computation(ctx)


def start_as_task(
    computation: Async[_T], cancellation_token: Optional[CancellationToken] = None
) -> Awaitable[_T]:
    """Executes a computation in the thread pool.

    If no cancellation token is provided then the default cancellation
    token is used.
    """
    tcs: TaskCompletionSource[_T] = TaskCompletionSource()

    def resolve(value: Optional[_T] = None) -> None:
        tcs.SetResult(value)

    def reject(error: Exception) -> None:
        tcs.SetException(error)

    def cancel(_: OperationCanceledError) -> None:
        tcs.SetCancelled()

    start_with_continuations(
        computation,
        resolve,
        reject,
        cancel,
        cancellation_token or default_cancellation_token,
    )
    return tcs.get_task()


def start_immediate(
    computation: Async[Any],
    cancellation_token: Optional[CancellationToken] = None,
) -> None:
    """Start computation immediately.

    Runs an asynchronous computation, starting immediately on the
    current operating system thread
    """
    return start_with_continuations(computation, cancellation_token=cancellation_token)


_executor: Optional[ThreadPoolExecutor] = None


def start(
    computation: Callable[[IAsyncContext[Any]], None],
    cancellation_token: Optional[CancellationToken] = None,
) -> None:
    """Starts the asynchronous computation.

    Starts the asynchronous computation in the thread pool. Do not await
    its result.

    If no cancellation token is provided then the default cancellation
    token is used."""
    global _executor

    def worker() -> None:
        start_immediate(computation, cancellation_token)

    if not _executor:
        _executor = ThreadPoolExecutor(max_workers=16)
    _executor.submit(worker)
    return


def run_synchronously(
    computation: Async[_T], cancellation_token: Optional[CancellationToken] = None
) -> _T:
    """Run computation synchronously.

    Runs an asynchronous computation and awaits its result on the
    calling thread. Propagates an exception should the computation yield
    one. This call is blocking.
    """

    async def runner() -> _T:
        return await start_as_task(computation, cancellation_token=cancellation_token)

    return asyncio.run(runner())


__all__ = [
    "await_task",
    "cancel",
    "cancel_after",
    "catch_async",
    "create_cancellation_token",
    "from_continuations",
    "ignore",
    "is_cancellation_requested",
    "sleep",
    "start",
    "start_immediate",
    "start_with_continuations",
]



"""Task handling.

This module implements .NET
[tasks](https://docs.microsoft.com/en-us/dotnet/standard/async-in-depth)
using Python async / await.
"""
import asyncio
from asyncio import AbstractEventLoop, Future
from typing import TYPE_CHECKING, Any, Awaitable, Generic, TypeVar

_T = TypeVar("_T")


class TaskCompletionSource(Generic[_T]):
    def __init__(self) -> None:
        self.loop: AbstractEventLoop = asyncio.get_event_loop()
        if TYPE_CHECKING:
            self.future: Future[_T] = self.loop.create_future()
        else:
            self.future: Future = self.loop.create_future()

    def SetResult(self, value: _T) -> None:
        """Set result.

        Transitions the underlying Task[TResult] into the
        RanToCompletion state.
        """

        def action():
            self.future.set_result(value)

        self.loop.call_soon_threadsafe(action)

    def SetCancelled(self) -> None:
        """Set cancelled.

        Transitions the underlying Task[TResult] into the Canceled
        state.
        """

        def action():
            self.future.cancel()

        self.loop.call_soon_threadsafe(action)

    def SetException(self, exception: Exception) -> None:
        """Set exception.

        Transitions the underlying Task[TResult] into the Faulted state
        and binds it to a specified exception.
        """

        def action():
            self.future.set_exception(exception)

        self.loop.call_soon_threadsafe(action)

    def get_task(self) -> Awaitable[_T]:
        return asyncio.ensure_future(self.future)


async def zero() -> None:
    return


async def from_result(value: _T) -> _T:
    return value


def get_awaiter(value: Awaitable[_T]) -> Awaitable[_T]:

    # Wrap awaitable in coroutine (so we can run it using create_task)
    async def get_value() -> _T:
        return await asyncio.ensure_future(value)

    return get_value()


def get_result(value: Awaitable[_T]) -> _T:
    """Get the result.

    Ends the wait for the completion of the asynchronous task.
    """

    async def runner() -> _T:
        return await value

    return asyncio.run(runner())


def start(computation: Awaitable[Any]) -> None:
    asyncio.create_task(computation)

    return None


__all__ = [
    "get_awaiter",
    "get_result",
    "from_result",
    "start",
    "TaskCompletionSource",
    "zero",
]



from decimal import Decimal
from typing import Any

from .types import FSharpRef


def from_zero() -> int:
    return 0


def from_one() -> int:
    return 1


def from_int32(x: int) -> int:
    return x


def from_int64(x: int) -> int:
    return x


def from_string(x: str) -> int:
    return int(x)


def op_addition(a: int, b: int) -> int:
    return a + b


def parse(value: Any) -> int:
    return int(value)


def to_string(value: int) -> str:
    return str(value)


def to_byte(value: int) -> int:
    return value


def to_sbyte(value: int) -> int:
    return value


def to_int16(value: int) -> int:
    return value


def to_uint16(value: int) -> int:
    return value


def to_int32(value: int) -> int:
    return value


def to_uint32(value: int) -> int:
    return value


def to_int64(value: int) -> int:
    return value


def to_uint64(value: int) -> int:
    return value


def to_single(value: int) -> float:
    return float(value)


def to_double(value: int) -> float:
    return float(value)


def to_decimal(value: int) -> Decimal:
    return Decimal(value)


def equals(a: int, b: int) -> bool:
    return a == b


def try_parse(string: str, defValue: FSharpRef[int]) -> bool:
    try:
        defValue.contents = parse(string)
        return True
    except Exception:
        return False


BigInteger = int

__all__ = [
    "BigInteger",
    "equals",
    "from_zero",
    "from_one",
    "from_int32",
    "from_int64",
    "from_string",
    "to_decimal",
    "to_double",
    "to_single",
    "to_uint64",
    "to_int64",
    "to_uint32",
    "to_int32",
    "to_uint16",
    "to_int16",
    "to_byte",
    "to_sbyte",
    "to_string",
    "op_addition",
    "parse",
    "try_parse",
]



from __future__ import annotations
from typing import (Optional, Any, TypeVar, Generic, List)
from ..fable_library.reflection import (TypeInfo, class_type, string_type, option_type, record_type)
from ..fable_library.string import (to_console, printf)
from ..fable_library.types import (Exception, Record)
from ..fable_library.util import equals

_B = TypeVar("_B")

_S = TypeVar("_S")

def expr_74() -> TypeInfo:
    return class_type("FsHafas.Client.Log", None, Log)


class Log:
    pass

Log_reflection = expr_74

def Log__ctor() -> Log:
    return Log()


def expr_75() -> TypeInfo:
    return class_type("FsHafas.Client.HafasError", None, HafasError, class_type("System.Exception"))


class HafasError(Exception):
    def __init__(self, code: str, msg: str) -> None:
        super().__init__(msg)
        self.code_004027 = code
    

HafasError_reflection = expr_75

def HafasError__ctor_Z384F8060(code: str, msg: str) -> HafasError:
    return HafasError(code, msg)


def expr_76() -> TypeInfo:
    return record_type("FsHafas.Client.Icon", [], Icon, lambda: [("type", string_type), ("title", option_type(string_type))])


class Icon(Record):
    def __init__(self, type: str, title: Optional[str]) -> None:
        super().__init__()
        self.type = type
        self.title = title
    

Icon_reflection = expr_76

def Log__cctor(__unit: Any=None) -> None:
    Log.debug = False


Log__cctor()

def Log_get_Debug() -> bool:
    return Log.debug


def Log_set_Debug_Z1FBCCD16(v: bool) -> None:
    Log.debug = v


def Log_Print(msg: str, o: Any=None) -> None:
    if Log.debug:
        to_console(printf("%s %A"))(msg)(o)
    


def HafasError__get_code(e: HafasError) -> str:
    return e.code_004027


def HafasError__get_isHafasError(e: HafasError) -> bool:
    return True


def expr_77(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return class_type("FsHafas.Client.IndexMap`2", [gen0, gen1], IndexMap_2)


class IndexMap_2(Generic[_B, _S]):
    def __init__(self, default_value: Optional[_B]=None) -> None:
        self.default_value = default_value
        self.dict = dict()
    

IndexMap_2_reflection = expr_77

def IndexMap_2__ctor_2B594(default_value: Optional[_B]=None) -> IndexMap_2[_B]:
    return IndexMap_2(default_value)


def IndexMap_2__get_Item_2B595(__: IndexMap_2[_S, _B], s: _S) -> _B:
    v : _B = __.dict.get(s)
    if not equals(v, None):
        return v
    
    else: 
        return __.default_value
    


def IndexMap_2__set_Item_541DA560(__: IndexMap_2[_S, _B], s: Any=None, b: Any=None) -> None:
    __.dict[s]=b


def IndexMap_2__get_Keys(__: IndexMap_2[_S, _B]) -> List[_S]:
    return list(__.dict)





from __future__ import annotations
from typing import (List, Optional, Any)
from ...fable_library.date import now
from ...fable_library.string import (to_text, printf)
from ...fs_hafas_python.context import (Context, Profile__get_parseJourneyLeg)
from ...fs_hafas_python.lib.transformations import (RawDep_FromRawStopL, RawArr_FromRawStopL, ToTrip_FromLeg)
from ...fs_hafas_python.types_hafas_client import (Leg, Trip)
from ...fs_hafas_python.types_raw_hafas_client import (RawJny, RawStop, RawSec)

def parse_trip(ctx: Context, j: RawJny) -> Trip:
    match_value : Optional[List[RawStop]] = j.stop_l
    (pattern_matching_result, stop_l_1) = (None, None)
    if match_value is not None:
        if len(match_value) > 1:
            pattern_matching_result = 0
            stop_l_1 = match_value
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        raw_sec_l : RawSec = RawSec("JNY", None, RawDep_FromRawStopL(stop_l_1[0]), RawArr_FromRawStopL(stop_l_1[len(stop_l_1) - 1]), j, None, None, None)
        date_1 : str
        match_value_1 : Optional[str] = j.date
        if match_value_1 is None:
            dt : Any = now()
            arg30 : int = (dt.day) or 0
            arg20 : int = (dt.month) or 0
            arg10 : int = (dt.year) or 0
            date_1 = to_text(printf("%04d%02d%02d"))(arg10)(arg20)(arg30)
        
        else: 
            date_1 = match_value_1
        
        leg : Leg = Profile__get_parseJourneyLeg(ctx.profile)(ctx)(raw_sec_l)(date_1)
        match_value_2 : Optional[str] = leg.trip_id
        if match_value_2 is not None:
            return ToTrip_FromLeg(match_value_2, leg)
        
        else: 
            raise Exception("parseTrip failed")
        
    
    elif pattern_matching_result == 1:
        raise Exception("parseTrip failed")
    





from __future__ import annotations
from typing import (Optional, List, Any, Tuple, TypeVar,
                    Tuple, Awaitable, Union, Callable, Dict)
from .fable_modules.fable_library.array import (map)
from .fable_modules.fable_library.task import TaskCompletionSource
from .fable_modules.fable_library.async_builder import (singleton, Async)
from .fable_modules.fable_library.util import IDisposable
from .fable_modules.fable_library.async_ import (
    start_with_continuations, start_as_task, default_cancellation_token)
from .fable_modules.fable_library.util import (
    create_atom, ignore, structural_hash, string_hash)
from .fable_modules.fable_library.string import (to_fail, printf, to_console)
from .fable_modules.fs_hafas_python.types_hafas_client import (Profile, IndexMap_2, Station, Stop, Location, ProductType, JourneysOptions, Journeys, RefreshJourneyOptions, Journey, StopOver, JourneysFromTripOptions, TripOptions, Trip, DeparturesArrivalsOptions,
                                                               Alternative, LocationsOptions, StopOptions, Location, NearByOptions, ReachableFromOptions, Duration, BoundingBox, RadarOptions, Movement, TripsByNameOptions, RemarksOptions, Warning, LinesOptions, Line, ServerOptions, ServerInfo)
from .fable_modules.fs_hafas_python.hafas_async_client import (HafasAsyncClient_productsOfMode, HafasAsyncClient, HafasAsyncClient__AsyncLocations, HafasAsyncClient__AsyncJourneys, HafasAsyncClient__AsyncJourneysFromTrip, HafasAsyncClient__AsyncRefreshJourney,
                                                               HafasAsyncClient__AsyncDepartures, HafasAsyncClient__AsyncArrivals, HafasAsyncClient__AsyncTripsByName, HafasAsyncClient__AsyncNearby, HafasAsyncClient__AsyncReachableFrom, HafasAsyncClient__AsyncRadar, HafasAsyncClient__AsyncStop, HafasAsyncClient__AsyncLines)
from .fable_modules.fs_hafas_python.context import (
    Profile as Profile_1)
from .fable_modules.fs_hafas_python.lib.transformations import (
    Default_LocationsOptions, Default_JourneysOptions, Default_Location, Default_DeparturesArrivalsOptions)

# todo: implement all methods of HafasAsyncClient


def _checkProfileType(profile: Any):
    # check, if profile is from fshafas.fable_modules.fs_hafas_profiles_python classes
    if not isinstance(profile, Profile_1):
        raise TypeError(
            "argument profile: type from fshafas.fable_modules.fs_hafas_profiles_python expected ")


class HafasClient(IDisposable):
    def __init__(self, profile: Profile) -> None:
        _checkProfileType(profile)

        self.profile = profile
        # pytype: disable=wrong-arg-types
        self.asyncClient = HafasAsyncClient(self.profile)

    async def _journeys(self, _from: Union[str, Station, Stop, Location], _to: Union[str, Station, Stop, Location], opt: Optional[JourneysOptions] = None) -> Journeys:
        def generate() -> Async[Journeys]:
            def bind(_arg1: Journeys) -> Async[Journeys]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncJourneys(self.asyncClient, _from, _to, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    async def journeys(self, _from: Union[str, Station, Stop, Location], _to: Union[str, Station, Stop, Location], opt: Optional[JourneysOptions] = None) -> Journeys:
        if opt is None:
            opt = Default_JourneysOptions

        if (isinstance(_from, str) and not _from.isdigit()):
            from_locs= await self.locations(_from, Default_LocationsOptions)
            if (len(from_locs) > 0 and from_locs[0].type == "stop"):
                _from= from_locs[0].id

        if (isinstance(_to, str) and not _to.isdigit()):
            to_locs= await self.locations(_to, Default_LocationsOptions)
            if (len(to_locs) > 0 and to_locs[0].type == "stop"):
                _to= to_locs[0].id

        return await self._journeys(_from, _to, opt)

    async def departures(self, name: Union[str, Station, Stop, Location], opt: Optional[DeparturesArrivalsOptions]=None) -> List[Alternative]:
        if (opt is not None and not isinstance(opt, DeparturesArrivalsOptions)):
            raise TypeError(
                "argument opt: type DeparturesArrivalsOptions expected")

        def generate() -> Async[List[Alternative]]:
            def bind(_arg1: List[Alternative]) -> Async[List[Alternative]]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncDepartures(self.asyncClient, name, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    async def arrivals(self, name: Union[str, Station, Stop, Location], opt: Optional[DeparturesArrivalsOptions]=None) -> List[Alternative]:
        if (opt is not None and not isinstance(opt, DeparturesArrivalsOptions)):
            raise TypeError(
                "argument opt: type DeparturesArrivalsOptions expected")

        def generate() -> Async[List[Alternative]]:
            def bind(_arg1: List[Alternative]) -> Async[List[Alternative]]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncArrivals(self.asyncClient, name, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    async def locations(self, name: str, opt: Optional[LocationsOptions]=None) -> List[Union[Station, Stop, Location]]:
        if opt is None:
            opt = Default_LocationsOptions

        if (not isinstance(name, str)):
            raise TypeError("argument name: type string expected")

        if (opt is not None and not isinstance(opt, LocationsOptions)):
            raise TypeError("argument opt: type LocationsOptions expected")

        def generate() -> Async[List[Any]]:
            def bind(_arg1: List[Any]) -> Async[List[Any]]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncLocations(self.asyncClient, name, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    async def nearby(self, l: Location, opt: Optional[NearByOptions]=None) -> List[Union[Station, Stop, Location]]:
        if (not isinstance(l, Location)):
            raise TypeError("argument l: type Location expected")

        if (opt is not None and not isinstance(opt, NearByOptions)):
            raise TypeError("argument opt: type NearByOptions expected")

        def generate() -> Async[List[Any]]:
            def bind(_arg1: List[Any]) -> Async[List[Any]]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncNearby(self.asyncClient, l, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    async def radar(self, rect: BoundingBox, opt: Optional[RadarOptions]=None) -> List[Movement]:
        if (not isinstance(rect, BoundingBox)):
            raise TypeError("argument rect: type BoundingBox expected")

        def generate() -> Async[List[Movement]]:
            def bind(_arg1: List[Movement]) -> Async[List[Movement]]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncRadar(self.asyncClient, rect, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    async def tripsByName(self, name: str, opt: Optional[TripsByNameOptions]=None) -> List[Trip]:
        if (not isinstance(name, str)):
            raise TypeError("argument name: type string expected")

        if (opt is not None and not isinstance(opt, TripsByNameOptions)):
            raise TypeError("argument opt: type TripsByNameOptions expected")

        def generate() -> Async[List[Any]]:
            def bind(_arg1: List[Any]) -> Async[List[Any]]:
                return singleton.Return(_arg1)

            return singleton.Bind(HafasAsyncClient__AsyncTripsByName(self.asyncClient, name, opt), bind)

        return await start_as_task(singleton.Delay(generate))

    def productsOfMode(self, profile: Profile, mode: str) -> IndexMap_2[str, bool]:
        _checkProfileType(profile)

        # pytype: disable=wrong-arg-types
        return HafasAsyncClient_productsOfMode(profile, mode)

    def Dispose(self) -> None:
        __: HafasClient= self
        __.asyncClient.Dispose()



from __future__ import annotations
from abc import abstractmethod
from typing import (Protocol, List, Optional)
from ...fable_library.list import FSharpList

class AST(Protocol):
    @property
    @abstractmethod
    def foo(self) -> int:
        ...
    

class mod(AST):
    pass

class expr(AST):
    pass

class Module(mod, AST):
    @property
    @abstractmethod
    def body(self) -> List[stmt]:
        ...
    

class Expression(mod, AST):
    @property
    @abstractmethod
    def body(self) -> expr:
        ...
    

class stmt(AST):
    pass

class FunctionDef(stmt, AST):
    @property
    @abstractmethod
    def args(self) -> arguments:
        ...
    
    @property
    @abstractmethod
    def body(self) -> List[stmt]:
        ...
    
    @property
    @abstractmethod
    def decorator_list(self) -> List[expr]:
        ...
    
    @property
    @abstractmethod
    def name(self) -> str:
        ...
    
    @property
    @abstractmethod
    def returns(self) -> Optional[expr]:
        ...
    

class ClassDef(stmt, AST):
    @property
    @abstractmethod
    def bases(self) -> List[expr]:
        ...
    
    @property
    @abstractmethod
    def body(self) -> List[stmt]:
        ...
    
    @property
    @abstractmethod
    def decorator_list(self) -> List[expr]:
        ...
    
    @property
    @abstractmethod
    def keywords(self) -> List[keyword]:
        ...
    
    @property
    @abstractmethod
    def name(self) -> str:
        ...
    

class Return(stmt, AST):
    @property
    @abstractmethod
    def value(self) -> Optional[expr]:
        ...
    

class Delete(stmt, AST):
    @property
    @abstractmethod
    def targets(self) -> List[expr]:
        ...
    

class Assign(stmt, AST):
    @property
    @abstractmethod
    def targets(self) -> List[expr]:
        ...
    
    @property
    @abstractmethod
    def value(self) -> expr:
        ...
    

class Import(stmt, AST):
    @property
    @abstractmethod
    def names(self) -> List[alias]:
        ...
    

class ImportFrom(stmt, AST):
    @property
    @abstractmethod
    def level(self) -> int:
        ...
    
    @property
    @abstractmethod
    def module(self) -> Optional[str]:
        ...
    
    @property
    @abstractmethod
    def names(self) -> List[alias]:
        ...
    

class If(stmt, AST):
    @property
    @abstractmethod
    def body(self) -> List[stmt]:
        ...
    
    @property
    @abstractmethod
    def orelse(self) -> List[stmt]:
        ...
    
    @property
    @abstractmethod
    def test(self) -> expr:
        ...
    

class arguments(AST):
    @property
    @abstractmethod
    def args(self) -> List[arg]:
        ...
    
    @property
    @abstractmethod
    def defaults(self) -> List[expr]:
        ...
    
    @property
    @abstractmethod
    def kw_defaults(self) -> FSharpList[Optional[expr]]:
        ...
    
    @property
    @abstractmethod
    def kwarg(self) -> Optional[arg]:
        ...
    
    @property
    @abstractmethod
    def kwonlyargs(self) -> List[arg]:
        ...
    
    @property
    @abstractmethod
    def posonlyargs(self) -> List[arg]:
        ...
    
    @property
    @abstractmethod
    def vararg(self) -> Optional[arg]:
        ...
    

class arg(AST):
    @property
    @abstractmethod
    def annotation(self) -> Optional[expr]:
        ...
    
    @property
    @abstractmethod
    def arg(self) -> str:
        ...
    

class keyword(AST):
    @property
    @abstractmethod
    def arg(self) -> Optional[str]:
        ...
    
    @property
    @abstractmethod
    def value(self) -> expr:
        ...
    

class alias(AST):
    @property
    @abstractmethod
    def asname(self) -> Optional[str]:
        ...
    
    @property
    @abstractmethod
    def name(self) -> str:
        ...
    

class IExports(Protocol):
    @abstractmethod
    def unparse(self, astObj: AST) -> str:
        ...
    
    @abstractmethod
    def walk(self, node: AST) -> List[AST]:
        ...
    




from typing import Any
from ...fable_library.string import (to_text, printf)

def format_date(dt: Any, pattern: str) -> str:
    if "yyyyMMdd" == pattern:
        arg30 : int = (dt.day) or 0
        arg20 : int = (dt.month) or 0
        arg10 : int = (dt.year) or 0
        return to_text(printf("%04d%02d%02d"))(arg10)(arg20)(arg30)
    
    else: 
        raise Exception("nyi")
    


def format_time(dt: Any, pattern: str) -> str:
    if "HHmm" == pattern:
        arg20 : int = (dt.minute) or 0
        arg10 : int = (dt.hour) or 0
        return to_text(printf("%02d%02d"))(arg10)(arg20)
    
    else: 
        raise Exception("nyi")
    





from __future__ import annotations
from typing import (Any, Optional, TypeVar, Callable)
from ...fable_library.date import now
from ...fable_library.option import (some, value as value_1)
from ...fable_library.util import round
from ...fs_hafas_python.context import Options
from ...fs_hafas_python.extra_types import IndexMap_2__ctor_2B594
from ...fs_hafas_python.types_hafas_client import (Leg, Trip, Station, Stop, Location, JourneysOptions, JourneysFromTripOptions, LocationsOptions, NearByOptions, TripOptions, LinesOptions, RemarksOptions, RefreshJourneyOptions, TripsByNameOptions, DeparturesArrivalsOptions, StopOptions, ReachableFromOptions, RadarOptions, ServerOptions, StopOver, Movement, Alternative, Journey, Journeys, Line, Warning, ServerInfo)
from ...fs_hafas_python.types_raw_hafas_client import (RawStop, RawDep, RawArr)

_A = TypeVar("_A")

_B = TypeVar("_B")

def Coordinate_toFloat(v: int) -> float:
    return v / 1000000


def Coordinate_fromFloat(x: float) -> int:
    return int(round(x * 1000000))


def RawDep_FromRawStopL(s: RawStop) -> RawDep:
    return RawDep(s.loc_x, s.idx, s.d_prod_x, s.d_platf_s, s.d_in_r, s.d_time_s, s.d_prog_type, None, s.d_tzoffset, None, s.d_time_r, s.d_cncl, s.d_pltf_s, s.d_platf_r, s.d_pltf_r)


def RawArr_FromRawStopL(s: RawStop) -> RawArr:
    return RawArr(s.loc_x, s.idx, s.a_platf_s, s.a_out_r, s.a_time_s, s.a_prog_type, s.a_tzoffset, None, s.a_time_r, s.a_cncl, s.a_pltf_s, s.a_platf_r, s.a_pltf_r, None)


def ToTrip_FromLeg(id: str, l: Leg) -> Trip:
    return Trip(id, l.origin, l.destination, l.departure, l.planned_departure, l.prognosed_arrival, l.departure_delay, l.departure_platform, l.prognosed_departure_platform, l.planned_departure_platform, l.arrival, l.planned_arrival, l.prognosed_departure, l.arrival_delay, l.arrival_platform, l.prognosed_arrival_platform, l.planned_arrival_platform, l.stopovers, l.schedule, l.price, l.operator, l.direction, l.line, l.reachable, l.cancelled, l.walking, l.load_factor, l.distance, l.public, l.transfer, l.cycle, l.alternatives, l.polyline, l.remarks)


def U2StationStop_FromU3StationStopLocation(u3: Any=None) -> Optional[Any]:
    if isinstance(u3, Station):
        return some(u3)
    
    elif isinstance(u3, Stop):
        return some(u3)
    
    else: 
        return None
    


def U2StationStop_FromSomeU3StationStopLocation(u3: Optional[Any]=None) -> Optional[Any]:
    (pattern_matching_result, s, s_1) = (None, None, None)
    if u3 is not None:
        if isinstance(value_1(u3), Station):
            pattern_matching_result = 0
            s = value_1(u3)
        
        elif isinstance(value_1(u3), Stop):
            pattern_matching_result = 1
            s_1 = value_1(u3)
        
        else: 
            pattern_matching_result = 2
        
    
    else: 
        pattern_matching_result = 2
    
    if pattern_matching_result == 0:
        return some(s)
    
    elif pattern_matching_result == 1:
        return some(s_1)
    
    elif pattern_matching_result == 2:
        return None
    


def U2StopLocation_FromU3StationStopLocation(u3: Any=None) -> Optional[Any]:
    if isinstance(u3, Stop):
        return some(u3)
    
    elif isinstance(u3, Location):
        return some(u3)
    
    else: 
        return None
    


def U2StopLocation_FromSomeU3StationStopLocation(u3: Optional[Any]=None) -> Optional[Any]:
    (pattern_matching_result, s, s_1) = (None, None, None)
    if u3 is not None:
        if isinstance(value_1(u3), Stop):
            pattern_matching_result = 0
            s = value_1(u3)
        
        elif isinstance(value_1(u3), Location):
            pattern_matching_result = 1
            s_1 = value_1(u3)
        
        else: 
            pattern_matching_result = 2
        
    
    else: 
        pattern_matching_result = 2
    
    if pattern_matching_result == 0:
        return some(s)
    
    elif pattern_matching_result == 1:
        return some(s_1)
    
    elif pattern_matching_result == 2:
        return None
    


def MergeOptions_getOptionValue(opt: Optional[_A], getter: Callable[[_A], Optional[_B]], default_value: _B) -> _B:
    if opt is None:
        return default_value
    
    else: 
        match_value : Optional[_B] = getter(value_1(opt))
        if match_value is None:
            return default_value
        
        else: 
            return value_1(match_value)
        
    


def MergeOptions_JourneysOptions(options: Options, opt: Optional[JourneysOptions]=None) -> Options:
    def arrow_303(v: JourneysOptions, options: Options=options, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v.remarks
    
    def arrow_304(v_1: JourneysOptions, options: Options=options, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v_1.stopovers
    
    def arrow_305(v_2: JourneysOptions, options: Options=options, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v_2.scheduled_days
    
    def arrow_306(v_3: JourneysOptions, options: Options=options, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v_3.first_class
    
    return Options(MergeOptions_getOptionValue(opt, arrow_303, options.remarks), MergeOptions_getOptionValue(opt, arrow_304, options.stopovers), options.polylines, MergeOptions_getOptionValue(opt, arrow_305, options.scheduled_days), options.sub_stops, options.entrances, options.lines_of_stops, MergeOptions_getOptionValue(opt, arrow_306, options.first_class))


def MergeOptions_JourneysFromTripOptions(options: Options, opt: Optional[JourneysFromTripOptions]=None) -> Options:
    def arrow_307(v: JourneysFromTripOptions, options: Options=options, opt: Optional[JourneysFromTripOptions]=opt) -> Optional[bool]:
        return v.stopovers
    
    return Options(options.remarks, MergeOptions_getOptionValue(opt, arrow_307, options.stopovers), options.polylines, options.scheduled_days, options.sub_stops, options.entrances, options.lines_of_stops, options.first_class)


def MergeOptions_LocationsOptions(options: Options, opt: Optional[LocationsOptions]=None) -> Options:
    def arrow_308(v: LocationsOptions, options: Options=options, opt: Optional[LocationsOptions]=opt) -> Optional[bool]:
        return v.lines_of_stops
    
    return Options(options.remarks, options.stopovers, options.polylines, options.scheduled_days, options.sub_stops, options.entrances, MergeOptions_getOptionValue(opt, arrow_308, options.lines_of_stops), options.first_class)


def MergeOptions_NearByOptions(options: Options, opt: Optional[NearByOptions]=None) -> Options:
    def arrow_309(v: NearByOptions, options: Options=options, opt: Optional[NearByOptions]=opt) -> Optional[bool]:
        return v.lines_of_stops
    
    return Options(options.remarks, options.stopovers, options.polylines, options.scheduled_days, options.sub_stops, options.entrances, MergeOptions_getOptionValue(opt, arrow_309, options.lines_of_stops), options.first_class)


Default_TripOptions : TripOptions = TripOptions(True, True, True, True, True, "de")

Default_LinesOptions : LinesOptions = LinesOptions("de")

Default_RemarksOptions : RemarksOptions = RemarksOptions(now(), None, 100, IndexMap_2__ctor_2B594(False), False, "de")

Default_LocationsOptions : LocationsOptions = LocationsOptions(True, 5, True, True, True, True, True, False, "de")

Default_RefreshJourneyOptions : RefreshJourneyOptions = RefreshJourneyOptions(True, False, False, None, None, None, "de")

Default_TripsByNameOptions : TripsByNameOptions = TripsByNameOptions(now(), None, None, None, None, None, None, None, None)

Default_NearByOptions : NearByOptions = NearByOptions(8, -1, None, True, IndexMap_2__ctor_2B594(False), None, None, None, "de")

Default_JourneysOptions : JourneysOptions = JourneysOptions(now(), None, None, None, 3, None, True, -1, 0, "none", False, IndexMap_2__ctor_2B594(False), False, False, True, True, True, "normal", True, "de", False, False, None, None, None)

Default_JourneysFromTripOptions : JourneysFromTripOptions = JourneysFromTripOptions(False, 0, "none", False, False, True, True, True, IndexMap_2__ctor_2B594(False))

Default_DeparturesArrivalsOptions : DeparturesArrivalsOptions = DeparturesArrivalsOptions(now(), None, None, 10, None, None, None, None, None, False, False, IndexMap_2__ctor_2B594(False), "de")

Default_StopOptions : StopOptions = StopOptions(None, None, None, None, "de")

Default_ReachableFromOptions : ReachableFromOptions = ReachableFromOptions(now(), 5, 20, IndexMap_2__ctor_2B594(False), None, None, None)

Default_RadarOptions : RadarOptions = RadarOptions(256, 3, IndexMap_2__ctor_2B594(False), 30, True, True, True, now())

Default_ServerOptions : ServerOptions = ServerOptions(True, "de")

Default_Location : Location = Location("location", None, None, None, None, None, None, None, None)

Default_Stop : Stop = Stop("stop", None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_Station : Station = Station("station", None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_Leg : Leg = Leg(None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_StopOver : StopOver = StopOver(None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_Trip : Trip = Trip("", None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_Movement : Movement = Movement(None, None, None, None, None, None, None)

Default_Alternative : Alternative = Alternative("", None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_Journey : Journey = Journey("journey", [], None, None, None, None, None)

Default_Journeys : Journeys = Journeys(None, None, None, None)

Default_Line : Line = Line("line", None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_Warning : Warning = Warning("warning", None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

Default_ServerInfo : ServerInfo = ServerInfo(None, None, None, None, None)




from typing import (Any, Optional, List)
from ..fable_library.option import some

def _007CNativeString_007C__007C(x: Any=None) -> Optional[str]:
    if isinstance(x, str):
        return x
    
    else: 
        return None
    


def _007CNativeBool_007C__007C(x: Any=None) -> Optional[bool]:
    if isinstance(x, bool):
        return x
    
    else: 
        return None
    


def _007CNativeNumber_007C__007C(x: Any=None) -> Optional[float]:
    if isinstance(x, int) or isinstance(x, float):
        return x
    
    else: 
        return None
    


def _007CNativeObject_007C__007C(x: Any=None) -> Optional[Any]:
    if isinstance(x, dict):
        return some(x)
    
    else: 
        return None
    


def _007CNull_007C__007C(x: Any=None) -> Optional[Any]:
    if x is None:
        return some(x)
    
    else: 
        return None
    


def _007CNativeArray_007C__007C(x: Any=None) -> Optional[List[Any]]:
    if isinstance(x, list):
        return x
    
    else: 
        return None
    





from abc import abstractmethod
from typing import (Tuple, Protocol)

class IExports(Protocol):
    @property
    @abstractmethod
    def altzone(self) -> int:
        ...
    
    @property
    @abstractmethod
    def daylight(self) -> int:
        ...
    
    @property
    @abstractmethod
    def timezone(self) -> int:
        ...
    
    @property
    @abstractmethod
    def tzname(self) -> Tuple[str, str]:
        ...
    
    @abstractmethod
    def monotonic(self) -> float:
        ...
    
    @abstractmethod
    def perf_counter(self) -> float:
        ...
    
    @abstractmethod
    def process_time(self) -> float:
        ...
    
    @abstractmethod
    def sleep(self, secs: float) -> None:
        ...
    
    @abstractmethod
    def time(self) -> float:
        ...
    




from typing import (
    Any,
    Awaitable,
    Callable,
    Iterable,
    Optional,
    Protocol,
    TypeVar,
    overload,
)

from .task import from_result, zero
from .util import IDisposable

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
TD = TypeVar("TD", bound=IDisposable)
U = TypeVar("U")


class Delayed(Protocol[T_co]):
    def __call__(self, __unit: Optional[None] = None) -> Awaitable[T_co]:
        ...


class TaskBuilder:
    def Bind(
        self, computation: Awaitable[T], binder: Callable[[T], Awaitable[U]]
    ) -> Awaitable[U]:
        async def bind() -> U:
            value = await computation
            return await binder(value)

        return bind()

    def Combine(
        self, computation1: Awaitable[None], computation2: Delayed[T]
    ) -> Awaitable[T]:
        return self.Bind(computation1, computation2)

    def Delay(self, generator: Callable[[], Awaitable[T]]) -> Delayed[T]:
        def deferred(_: Any = None) -> Awaitable[T]:
            # print("Delay: deferred: ", generator)
            return generator()

        return deferred

    def For(
        self, sequence: Iterable[T], body: Callable[[T], Awaitable[U]]
    ) -> Awaitable[U]:
        done = False
        it = iter(sequence)
        try:
            cur = next(it)
        except StopIteration:
            done = True

        def delay():
            nonlocal cur, done
            res = body(cur)
            try:
                cur = next(it)
            except StopIteration:
                done = True
            return res

        return self.While(lambda: not done, self.Delay(delay))

    @overload
    def Return(self) -> Awaitable[None]:
        ...

    @overload
    def Return(self, value: T) -> Awaitable[T]:
        ...

    def Return(self, value: Optional[T] = None) -> Awaitable[Optional[T]]:
        return from_result(value)

    def ReturnFrom(self, computation: Awaitable[T]) -> Awaitable[T]:
        return computation

    def TryFinally(
        self, computation: Delayed[T], compensation: Callable[[], None]
    ) -> Awaitable[T]:
        async def try_finally() -> T:
            try:
                t = await computation()
            finally:
                compensation()
            return t

        return try_finally()

    def TryWith(
        self, computation: Delayed[T], catchHandler: Callable[[Any], Awaitable[T]]
    ) -> Awaitable[T]:
        async def try_with() -> T:
            try:
                t = await computation()
            except Exception as exn:
                t = await catchHandler(exn)
            return t

        return try_with()

    def Using(self, resource: TD, binder: Callable[[TD], Awaitable[U]]) -> Awaitable[U]:
        return self.TryFinally(
            self.Delay(lambda: binder(resource)), lambda: resource.Dispose()
        )

    @overload
    def While(
        self, guard: Callable[[], bool], computation: Delayed[None]
    ) -> Awaitable[None]:
        ...

    @overload
    def While(self, guard: Callable[[], bool], computation: Delayed[T]) -> Awaitable[T]:  # type: ignore
        ...

    def While(
        self, guard: Callable[[], bool], computation: Delayed[Any]
    ) -> Awaitable[Any]:
        if guard():
            return self.Bind(computation(), lambda _: self.While(guard, computation))
        else:
            return self.Return()

    def Zero(self) -> Awaitable[None]:
        return zero()

    def Run(self, computation: Delayed[T]) -> Awaitable[T]:
        # Make sure we don't execute computation right now, so wrap in a coroutine.
        async def run() -> T:
            return await computation()

        return run()


task = TaskBuilder

__all__ = ["task"]



from __future__ import annotations
from typing import (Any, Optional)
from ...fable_library.list import (FSharpList, last)
from ...fable_library.reg_exp import replace
from ...fable_library.string import (substring, to_console, printf)
from ...fable_simple_json_python.json_converter import Convert_fromJson
from ...fable_simple_json_python.simple_json import (SimpleJson_mapKeysByPath, SimpleJson_parseNative)
from ...fable_simple_json_python.type_info_converter import create_type_info
from ...fs_hafas_python.types_raw_hafas_client import (RawResponse_reflection, RawResponse)

def dashify(separator: str, input: str) -> str:
    def arrow_297(m: Any, separator: str=separator, input: str=input) -> str:
        return m[0].lower() if (len(m[0]) == 1) else ((substring(m[0], 0, 1) + separator) + substring(m[0], 1, 1).lower())
    
    return replace(input, "[a-z]?[A-Z]", arrow_297)


def to_dashed(xs: FSharpList[str]) -> Optional[str]:
    return dashify("_", last(xs))


def decode(input: str) -> RawResponse:
    try: 
        def arrow_298(xs: FSharpList[str]) -> Optional[str]:
            return to_dashed(xs)
        
        return Convert_fromJson(SimpleJson_mapKeysByPath(arrow_298, SimpleJson_parseNative(input)), create_type_info(RawResponse_reflection()))
    
    except Exception as e:
        arg10 : str = str(e)
        to_console(printf("error decode: %s"))(arg10)
        raise Exception(str(e))
    





from typing import Tuple

from .types import FSharpRef


def get_range(unsigned: bool, bitsize: int) -> Tuple[int, int]:
    if bitsize == 8:
        return (0, 255) if unsigned else (-128, 127)
    if bitsize == 16:
        return (0, 65535) if unsigned else (-32768, 32767)
    if bitsize == 32:
        return (0, 4294967295) if unsigned else (-2147483648, 2147483647)
    raise ValueError("Invalid bit size.")


AllowHexSpecifier = 0x00000200


def parse(string: str, style, unsigned, bitsize, radix: int = 10) -> int:
    # const res = isValid(str, style, radix);
    if style & AllowHexSpecifier or string.startswith("0x"):
        radix = 16
    elif string.startswith("0b"):
        radix = 2
    elif string.startswith("0o"):
        radix = 8

    try:
        v = int(string, base=radix)
    except Exception:
        raise ValueError("Input string was not in a correct format.")

    (umin, umax) = get_range(True, bitsize)
    if not unsigned and radix != 10 and v >= umin and v <= umax:
        mask = 1 << (bitsize - 1)
        if v & mask:  # Test if negative
            v = v - (mask << 1)

    (min, max) = get_range(unsigned, bitsize)
    if v >= min and v <= max:
        return v

    raise ValueError("Input string was not in a correct format.")


def try_parse(
    string: str, style: int, unsigned: bool, bitsize: int, defValue: FSharpRef[int]
) -> bool:
    try:
        defValue.contents = parse(string, style, unsigned, bitsize)
        return True
    except Exception:
        return False


def op_unary_negation_int8(x):
    return x if x == -128 else -x


def op_unary_negation_int16(x):
    return x if x == -32768 else -x


def op_unary_negation_int32(x):
    return x if x == -2147483648 else -x



from __future__ import annotations
from typing import (Optional, Any, Tuple)
from ...fable_library.date import op_subtraction
from ...fable_library.option import (to_array, default_arg)
from ...fable_library.seq import exists
from ...fable_library.time_span import total_seconds
from ...fable_library.util import round
from ...fs_hafas_python.context import (ParsedWhen, Context)
from .date_time import parse_date_time_ex

default_when : ParsedWhen = ParsedWhen(None, None, None, None)

def parse_when(ctx: Context, date: str, time_s: Optional[str]=None, time_r: Optional[str]=None, tz_offset: Optional[int]=None, cncl: Optional[bool]=None) -> ParsedWhen:
    pattern_input : Tuple[Optional[Any], Optional[str]] = parse_date_time_ex(ctx.profile, date, time_s, tz_offset)
    str_planned : Optional[str] = pattern_input[1]
    pattern_input_1 : Tuple[Optional[Any], Optional[str]] = parse_date_time_ex(ctx.profile, date, time_r, tz_offset)
    str_prognosed : Optional[str] = pattern_input_1[1]
    delay : Optional[int]
    match_value : Tuple[Optional[Any], Optional[Any]] = (pattern_input_1[0], pattern_input[0])
    (pattern_matching_result, planned, prognosed) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            planned = match_value[1]
            prognosed = match_value[0]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def arrow_357(ctx: Context=ctx, date: str=date, time_s: Optional[str]=time_s, time_r: Optional[str]=time_r, tz_offset: Optional[int]=tz_offset, cncl: Optional[bool]=cncl) -> float:
            copy_of_struct : Any = op_subtraction(prognosed, planned)
            return total_seconds(copy_of_struct)
        
        delay = int(round(arrow_357()))
    
    elif pattern_matching_result == 1:
        delay = None
    
    def predicate(x: bool, ctx: Context=ctx, date: str=date, time_s: Optional[str]=time_s, time_r: Optional[str]=time_r, tz_offset: Optional[int]=tz_offset, cncl: Optional[bool]=cncl) -> bool:
        return x
    
    if exists(predicate, to_array(cncl)):
        return ParsedWhen(None, str_planned, str_prognosed, delay)
    
    else: 
        return ParsedWhen(default_arg(str_prognosed, str_planned), str_planned, None, delay)
    





from typing import Any, Callable, Generic, List, Optional, TypeVar, Union, cast

_T = TypeVar("_T")
_U = TypeVar("_U")
_V = TypeVar("_V")
_W = TypeVar("_W")


class Some(Generic[_T]):
    def __init__(self, value: _T):
        self.value = value

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True

        if other is None:
            return False

        if self.value == other.value:
            return True

        return False

    def __str__(self):
        return f"Some {self.value}"

    def __repr__(self):
        return str(self)


Option = Union[Some[_T], _T, None]


def default_arg(opt: Option[_T], default_value: _T) -> _T:
    return value(opt) if opt is not None else default_value


def default_arg_with(opt: Option[_T], def_thunk: Callable[[], _T]) -> _T:
    return value(opt) if opt is not None else def_thunk()


def filter(predicate: Callable[[_T], bool], opt: Option[_T]) -> Option[_T]:
    if opt is not None:
        return opt if predicate(value(opt)) else None
    return opt


def map(mapping: Callable[[_T], _U], opt: Option[_T]) -> Option[_U]:
    return some(mapping(value(opt))) if opt is not None else None


def map2(
    mapping: Callable[[_T, _U], _V], opt1: Option[_T], opt2: Option[_U]
) -> Option[_V]:
    return (
        mapping(value(opt1), value(opt2))
        if (opt1 is not None and opt2 is not None)
        else None
    )


def map3(
    mapping: Callable[[_T, _U, _V], _W],
    opt1: Option[_T],
    opt2: Option[_U],
    opt3: Option[_V],
) -> Option[_W]:
    return (
        mapping(value(opt1), value(opt2), value(opt3))
        if (opt1 is not None and opt2 is not None and opt3 is not None)
        else None
    )


def some(x: _T) -> Option[_T]:
    return Some(x) if x is None or isinstance(x, Some) else x  # type: ignore


def value(x: Option[_T]) -> _T:
    if x is None:
        raise Exception("Option has no value")
    elif isinstance(x, Some):
        y = cast(Some[_T], x)
        return y.value
    return x


def of_nullable(x: Optional[_T]) -> Option[_T]:
    return x


def to_nullable(x: Option[_T]) -> Optional[_T]:
    return None if x is None else value(x)


def flatten(x: Option[Option[_T]]) -> Option[_T]:
    return x if x is None else value(x)


def to_array(opt: Option[_T]) -> List[_T]:
    return [] if opt is None else [value(opt)]


def bind(binder: Callable[[_T], Option[_U]], opt: Option[_T]) -> Option[_U]:
    return binder(value(opt)) if opt is not None else None


__all__ = [
    "bind",
    "default_arg",
    "default_arg_with",
    "flatten",
    "map",
    "map2",
    "map3",
    "of_nullable",
    "Option",
    "some",
    "Some",
    "to_array",
    "to_nullable",
    "value",
]



from __future__ import annotations
from typing import (TypeVar, Any, Generic, Optional, List)
from .array import (fill, copy_to, initialize)
from .reflection import (TypeInfo, class_type)
from .seq import (delay, enumerate_while, append, singleton, empty, to_array)
from .types import FSharpRef
from .util import (compare, IComparer, equals, structural_hash, IEqualityComparer, get_enumerator, IEnumerable, IEnumerator, to_iterator, max, compare_primitives)

_T = TypeVar("_T")

_T_ = TypeVar("_T_")

def expr_10(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.Comparer`1", [gen0], Comparer_1)


class Comparer_1(Generic[_T]):
    def Compare(self, x: _T, y: _T) -> int:
        return compare(x, y)
    

Comparer_1_reflection = expr_10

def Comparer_1__ctor() -> Comparer_1[Any]:
    return Comparer_1()


def Comparer_1_get_Default() -> IComparer[_T]:
    class ObjectExpr11(IComparer[_T_]):
        def Compare(self, x: _T_, y: _T_) -> int:
            return compare(x, y)
        
    return ObjectExpr11()


def expr_12(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.EqualityComparer`1", [gen0], EqualityComparer_1)


class EqualityComparer_1(Generic[_T]):
    def __eq__(self, x: _T, y: _T) -> bool:
        return equals(x, y)
    
    def GetHashCode(self, x: Optional[_T]=None) -> int:
        return structural_hash(x)
    

EqualityComparer_1_reflection = expr_12

def EqualityComparer_1__ctor() -> EqualityComparer_1[Any]:
    return EqualityComparer_1()


def EqualityComparer_1_get_Default() -> IEqualityComparer[Any]:
    class ObjectExpr13(IEqualityComparer[Any]):
        def Equals(self, x: _T_, y: _T_) -> bool:
            return equals(x, y)
        
        def GetHashCode(self, x_1: Optional[_T_]=None) -> int:
            return structural_hash(x_1)
        
    return ObjectExpr13()


def expr_18(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.Stack`1", [gen0], Stack_1)


class Stack_1(Generic[_T]):
    def __init__(self, initial_contents: List[_T], initial_count: int) -> None:
        self.contents = initial_contents
        self.count = initial_count or 0
    
    def GetEnumerator(self) -> IEnumerator[_T]:
        this : Stack_1[_T] = self
        def arrow_17(__unit: Any=None) -> IEnumerable[_T]:
            index : int = (this.count - 1) or 0
            def arrow_14(__unit: Any=None) -> bool:
                return index >= 0
            
            def arrow_16(__unit: Any=None) -> IEnumerable[_T]:
                def arrow_15(__unit: Any=None) -> IEnumerable[_T]:
                    nonlocal index
                    index = (index - 1) or 0
                    return empty()
                
                return append(singleton(this.contents[index]), delay(arrow_15))
            
            return enumerate_while(arrow_14, delay(arrow_16))
        
        return get_enumerator(delay(arrow_17))
    
    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())
    
    def System_Collections_IEnumerable_GetEnumerator(self) -> IEnumerator[Any]:
        this : Stack_1[_T] = self
        return get_enumerator(this)
    

Stack_1_reflection = expr_18

def Stack_1__ctor_Z2E171D71(initial_contents: List[_T], initial_count: int) -> Stack_1[_T]:
    return Stack_1(initial_contents, initial_count)


def Stack_1__ctor_Z524259A4(initial_capacity: int) -> Stack_1[_T]:
    return Stack_1__ctor_Z2E171D71(fill([0] * initial_capacity, 0, initial_capacity, None), 0)


def Stack_1__ctor() -> Stack_1[_T]:
    return Stack_1__ctor_Z524259A4(4)


def Stack_1__ctor_BB573A(xs: IEnumerable[_T]) -> Stack_1[_T]:
    arr : List[_T] = list(xs)
    return Stack_1__ctor_Z2E171D71(arr, len(arr))


def Stack_1__Ensure_Z524259A4(this: Stack_1[_T], new_size: int) -> None:
    old_size : int = len(this.contents) or 0
    if new_size > old_size:
        old : List[_T] = this.contents
        def arrow_20(x: int, y: int, this: Stack_1[_T]=this, new_size: int=new_size) -> int:
            return compare_primitives(x, y)
        
        def arrow_21(x: int, y: int, this: Stack_1[_T]=this, new_size: int=new_size) -> int:
            return compare_primitives(x, y)
        
        this.contents = fill([0] * max(arrow_20, new_size, old_size * 2), 0, max(arrow_21, new_size, old_size * 2), None)
        copy_to(old, 0, this.contents, 0, this.count)
    


def Stack_1__get_Count(this: Stack_1[_T]) -> int:
    return this.count


def Stack_1__Pop(this: Stack_1[_T]) -> _T:
    this.count = (this.count - 1) or 0
    return this.contents[this.count]


def Stack_1__Peek(this: Stack_1[_T]) -> _T:
    return this.contents[this.count - 1]


def Stack_1__Contains_2B595(this: Stack_1[_T], x: _T) -> bool:
    found : bool = False
    i : int = 0
    while (not found) if (i < this.count) else False:
        if equals(x, this.contents[i]):
            found = True
        
        else: 
            i = (i + 1) or 0
        
    return found


def Stack_1__TryPeek_1F3DB691(this: Stack_1[_T], result: FSharpRef[_T]) -> bool:
    if this.count > 0:
        result.contents = Stack_1__Peek(this)
        return True
    
    else: 
        return False
    


def Stack_1__TryPop_1F3DB691(this: Stack_1[_T], result: FSharpRef[_T]) -> bool:
    if this.count > 0:
        result.contents = Stack_1__Pop(this)
        return True
    
    else: 
        return False
    


def Stack_1__Push_2B595(this: Stack_1[_T], x: _T) -> None:
    Stack_1__Ensure_Z524259A4(this, this.count + 1)
    this.contents[this.count] = x
    this.count = (this.count + 1) or 0


def Stack_1__Clear(this: Stack_1[_T]) -> None:
    this.count = 0
    fill(this.contents, 0, len(this.contents), None)


def Stack_1__TrimExcess(this: Stack_1[_T]) -> None:
    if (this.count / len(this.contents)) > 0.9:
        Stack_1__Ensure_Z524259A4(this, this.count)
    


def Stack_1__ToArray(this: Stack_1[_T]) -> List[_T]:
    def arrow_22(i: int, this: Stack_1[_T]=this) -> _T:
        return this.contents[(this.count - 1) - i]
    
    return initialize(this.count, arrow_22, None)


def expr_23(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.Queue`1", [gen0], Queue_1)


class Queue_1(Generic[_T]):
    def __init__(self, initial_contents: List[_T], initial_count: int) -> None:
        self.contents = initial_contents
        self.count = initial_count or 0
        self.head = 0
        self.tail = initial_count or 0
    
    def GetEnumerator(self) -> IEnumerator[_T]:
        this : Queue_1[_T] = self
        return get_enumerator(Queue_1__toSeq(this))
    
    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())
    
    def System_Collections_IEnumerable_GetEnumerator(self) -> IEnumerator[Any]:
        this : Queue_1[_T] = self
        return get_enumerator(this)
    

Queue_1_reflection = expr_23

def Queue_1__ctor_Z2E171D71(initial_contents: List[_T], initial_count: int) -> Queue_1[_T]:
    return Queue_1(initial_contents, initial_count)


def Queue_1__ctor_Z524259A4(initial_capacity: int) -> Queue_1[_T]:
    if initial_capacity < 0:
        raise Exception("capacity is less than 0")
    
    return Queue_1__ctor_Z2E171D71(fill([0] * initial_capacity, 0, initial_capacity, None), 0)


def Queue_1__ctor() -> Queue_1[_T]:
    return Queue_1__ctor_Z524259A4(4)


def Queue_1__ctor_BB573A(xs: IEnumerable[_T]) -> Queue_1[_T]:
    arr : List[_T] = list(xs)
    return Queue_1__ctor_Z2E171D71(arr, len(arr))


def Queue_1__get_Count(_: Queue_1[_T]) -> int:
    return _.count


def Queue_1__Enqueue_2B595(_: Queue_1[_T], value: _T) -> None:
    if _.count == Queue_1__size(_):
        Queue_1__ensure_Z524259A4(_, _.count + 1)
    
    _.contents[_.tail] = value
    _.tail = ((_.tail + 1) % Queue_1__size(_)) or 0
    _.count = (_.count + 1) or 0


def Queue_1__Dequeue(_: Queue_1[_T]) -> _T:
    if _.count == 0:
        raise Exception("Queue is empty")
    
    value : _T = _.contents[_.head]
    _.head = ((_.head + 1) % Queue_1__size(_)) or 0
    _.count = (_.count - 1) or 0
    return value


def Queue_1__Peek(_: Queue_1[_T]) -> _T:
    if _.count == 0:
        raise Exception("Queue is empty")
    
    return _.contents[_.head]


def Queue_1__TryDequeue_1F3DB691(this: Queue_1[_T], result: FSharpRef[_T]) -> bool:
    if this.count == 0:
        return False
    
    else: 
        result.contents = Queue_1__Dequeue(this)
        return True
    


def Queue_1__TryPeek_1F3DB691(this: Queue_1[_T], result: FSharpRef[_T]) -> bool:
    if this.count == 0:
        return False
    
    else: 
        result.contents = Queue_1__Peek(this)
        return True
    


def Queue_1__Contains_2B595(this: Queue_1[_T], x: _T) -> bool:
    found : bool = False
    i : int = 0
    while (not found) if (i < this.count) else False:
        if equals(x, this.contents[Queue_1__toIndex_Z524259A4(this, i)]):
            found = True
        
        else: 
            i = (i + 1) or 0
        
    return found


def Queue_1__Clear(this: Queue_1[_T]) -> None:
    this.count = 0
    this.head = 0
    this.tail = 0
    fill(this.contents, 0, Queue_1__size(this), None)


def Queue_1__TrimExcess(this: Queue_1[_T]) -> None:
    if (this.count / len(this.contents)) > 0.9:
        Queue_1__ensure_Z524259A4(this, this.count)
    


def Queue_1__ToArray(this: Queue_1[_T]) -> List[_T]:
    return to_array(Queue_1__toSeq(this))


def Queue_1__CopyTo_Z2E171D71(this: Queue_1[_T], target: List[_T], start: int) -> None:
    i : int = start or 0
    with get_enumerator(Queue_1__toSeq(this)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            item : _T = enumerator.System_Collections_Generic_IEnumerator_00601_get_Current()
            target[i] = item
            i = (i + 1) or 0


def Queue_1__size(this: Queue_1[_T]) -> int:
    return len(this.contents)


def Queue_1__toIndex_Z524259A4(this: Queue_1[_T], i: int) -> int:
    return (this.head + i) % Queue_1__size(this)


def Queue_1__ensure_Z524259A4(this: Queue_1[_T], required_size: int) -> None:
    new_buffer : List[_T] = fill([0] * required_size, 0, required_size, None)
    if this.head < this.tail:
        copy_to(this.contents, this.head, new_buffer, 0, this.count)
    
    else: 
        copy_to(this.contents, this.head, new_buffer, 0, Queue_1__size(this) - this.head)
        copy_to(this.contents, 0, new_buffer, Queue_1__size(this) - this.head, this.tail)
    
    this.head = 0
    this.tail = this.count or 0
    this.contents = new_buffer


def Queue_1__toSeq(this: Queue_1[_T]) -> IEnumerable[_T]:
    def arrow_27(this: Queue_1[_T]=this) -> IEnumerable[_T]:
        i : int = 0
        def arrow_24(__unit: Any=None) -> bool:
            return i < this.count
        
        def arrow_26(__unit: Any=None) -> IEnumerable[_T]:
            def arrow_25(__unit: Any=None) -> IEnumerable[_T]:
                nonlocal i
                i = (i + 1) or 0
                return empty()
            
            return append(singleton(this.contents[Queue_1__toIndex_Z524259A4(this, i)]), delay(arrow_25))
        
        return enumerate_while(arrow_24, delay(arrow_26))
    
    return delay(arrow_27)


